from .validate import ValidationException

def get_system_serials_of_attached_boxes(only_registered=False):
    from ..utils.queries import get_uniq_list_of_attached_boxes__by_serial
    infiboxes = get_uniq_list_of_attached_boxes__by_serial()
    return [(box.get_vendor().get_system_serial(),
             box.get_vendor().get_system_name()) for box in infiboxes
            if ((not only_registered) or (box.get_vendor().get_host_id() > 0))]

def get_list_of_supported_filesystems():
    """:returns: a platform-specific list of supported file systems.
    The first filesystem in the list will be treated as the default.
    On Unix, ext3 shall be the default"""
    # TODO: implement this for real
    from infi.storagemodel import get_storage_model
    model = get_storage_model()
    mount = model.get_mount_manager()
    filesystems = mount.get_creatable_file_systems()
    names = set(filesystem.get_name() for filesystem in filesystems)
    names.discard('minix')
    names.discard('reiserfs')
    return list(names)

def get_default_filesystem_choice(choices):
    return 'ext3' if 'ext3' in choices else 'NTFS' if 'NTFS' in choices else choices[0]

def _volume_exists(system_object, volume_name):
    volume = system_object.volumes.safe_get(name=volume_name)
    return volume is not None

def volume_exists(system_serial, volume_name):
    from ..utils.queries import get_system_object_by_serial
    system = get_system_object_by_serial(system_serial)
    return _volume_exists(system, volume_name)

def get_suggested_volume_name(box_id):
    from time import strftime
    from ..utils.queries import get_system_object_and_host

    system_object, host = get_system_object_and_host(box_id)
    hostname = host.get_name()

    suggested_volume_name = "{}-vol-{}".format(hostname, strftime("%Y%m%d-%H%M"))
    i = 0
    while _volume_exists(system_object, suggested_volume_name):
        i += 1
        suggested_volume_name = "{}-vol-{}-{}".format(hostname, strftime("%Y%m%d-%H%M"), i)
    return suggested_volume_name

def get_next_available_drive_letter(volume_name):
    from infi.storagemodel import get_storage_model
    available_letters = get_storage_model().get_mount_manager().get_available_drive_letters()
    return available_letters[0] if available_letters else \
        r"C:\mounts\{}".format(volume_name)

def mount_dir_exists_and_in_use(path):
    from os.path import exists, ismount
    from os import listdir
    return exists(path) and (ismount(path) or listdir(path))

def get_suggested_mount_point(volume_name):
    """:returns: the next-available drive-letter on Windows, and mount path on posix"""
    from os import name
    if name != "nt":
        suggested_mount_point = "/mnt/{}".format(volume_name)
        i = 0
        while mount_dir_exists_and_in_use(suggested_mount_point):
            i += 1
            suggested_mount_point = "/mnt/{}-{}".format(volume_name, i)
        return suggested_mount_point
    else:
        return get_next_available_drive_letter(volume_name)

def get_available_pools(system_serial):
    from ..utils.queries import get_system_object_by_serial, get_pools
    system = get_system_object_by_serial(system_serial)
    return [pool.get_name() for pool in get_pools(system)]

def _autodetect_with_options(arguments, key, options, custom_message=None):
    if len(options) == 1:
        arguments[key] = str(options[0])
        return
    if options:
        msg = "Could not autodetect {}. Choices are: {}"
        msg_fmt = msg.format(key.replace('--', ''), ', '.join(options))
    else:
        msg = custom_message or "No {0}s available"
        msg_fmt = msg.format(key.replace('--', ''))
    raise ValidationException(msg_fmt)

def _autodetect_system(arguments, only_registered=False):
    options = get_system_serials_of_attached_boxes(only_registered)
    if len(options) == 1:
        serial, name = options[0]
        arguments["--system"] = str(serial)
        return
    if not options:
        msg = "No{} systems available".format(" registered" if only_registered else "")
    else:
        options = ", ".join([", ".join([str(serial), name]) for serial, name in options])
        msg = "Could not autodetect system. Choices are: {}".format(options)
    raise ValidationException(msg)

def _autodetect_pool(arguments):
    from .validate import _validate_system
    _validate_system(arguments, True)
    options = get_available_pools(int(arguments["--system"]))
    _autodetect_with_options(arguments, "--pool", options)

def _autodetect_hostname(arguments):
    from infi.vendata.powertools.utils.queries import get_os_hostname
    arguments["--hostname"] = get_os_hostname()

def _autodetect_mount(arguments):
    from .validate import _validate_name
    _validate_name(arguments)
    arguments["--mount"] = get_suggested_mount_point(arguments["--name"])

def _autodetect_name(arguments):
    arguments["--name"] = get_suggested_volume_name(arguments["--system"])

def _autodetect_filesystem(arguments):
    arguments["--filesystem"] = get_default_filesystem_choice(get_list_of_supported_filesystems())

def autodetect(arguments, *args):
    for arg in args:
        if arguments["--" + arg] in ["autodetect", "autogenerate"]:
            autodetector = globals().get("_autodetect_" + arg)
            autodetector(arguments)
