import docopt
from infi.os_info import get_platform_string
from ..utils.mock_environment import consider_mock_environment
from ..utils.logging import debug_program, suppress_warnings, log_usage
from ..rescan.engine import do_rescan
from ..utils import init_colors
from .validate import validate, convert_size, get_storagemodel_device
from .autodetect import autodetect
from ..settings.runner import ReadinessRunner
from infi.credentials_store import Credentials


class RescanFailed(Exception):
    pass


def get_size_unit(arguments):
    size_unit = arguments['--units']
    SIZE_OPTIONS = ['GB', 'TB', 'GiB', 'TiB']
    if size_unit not in SIZE_OPTIONS:
        raise docopt.DocoptExit('Invalid unit size. Options are: ' + ', '.join(SIZE_OPTIONS))
    return size_unit


def run_settings_check(arguments):
    runner = ReadinessRunner(arguments['--fix'], arguments['--auto-fix'], arguments['--with-qr-codes'])
    return runner.run()


def run_connectivity_list(arguments):
    from ..connectivity.engine import print_output
    print_output(arguments['--json'])
    return 0


def run_pool_list(arguments):
    from ..pool_list.engine import run
    if arguments["--system"] == "autodetect":
        system = None
    else:
        validate(arguments, "system")
        system = int(arguments["--system"])
    size_unit = get_size_unit(arguments)
    should_do_rescan = arguments["--rescan"]
    run(system, size_unit, should_do_rescan)
    return 0


def run_volume_list(arguments):
    from ..volume_list.engine import run
    strategy = "brief"
    for option in ["summary", "json", "detailed"]:
        if arguments["--" + option]:
            strategy = option
    size_unit = get_size_unit(arguments)
    should_do_rescan = arguments["--rescan"]
    run(strategy, size_unit, should_do_rescan)
    return 0


def are_you_sure(arguments, get_workout):
    from infi.vendata.powertools.volume_resize.engine import ResizeError
    if not arguments["--yes"]:
        try:
            print get_workout(arguments)
            print "If this is absolutely what you want, pass the --yes flag in the command-line.\naborting."
        except ResizeError as e:
            print e
        raise SystemExit()


def is_filesystem_support_disabled():
    from ..utils import is_veritas_multipathing_installed
    return is_veritas_multipathing_installed() or "solaris" in get_platform_string() or "aix" in get_platform_string()


def prepare_provision_arguments(arguments):
    from .autodetect import _autodetect_system
    from .validate import _validate_system, _validate_filesystem_size, _validate_pool_size
    if is_filesystem_support_disabled():
        arguments["--no-filesystem"] = True
        arguments["--no-mount"] = True
        arguments["--mount"] = "-"
        arguments["--filesystem"] = "-"
    if arguments["--system"] == "autodetect":
        _autodetect_system(arguments, True)
    _validate_system(arguments, True)
    autodetect(arguments, "pool", "name", "mount", "filesystem")
    need_validation = ["name", "thin", "pool", "lun"]
    if not arguments["--no-filesystem"]:
        need_validation.append("filesystem")
        if not arguments["--no-mount"]:
            need_validation.append("mount")
    validate(arguments, *need_validation)
    arguments["<size>"] = convert_size(arguments["<size>"])
    _validate_pool_size(arguments)
    if not arguments['--no-filesystem']:
        _validate_filesystem_size(arguments["--filesystem"], arguments["<size>"])


def assert_readiness_for_provisioning():
    readiness = ReadinessRunner.load_previous_result()
    if readiness is None:
        print 'This host is not ready to perform volume operations, please run "infinihost settings check"'
        raise SystemExit(1)
    elif not readiness['provisioning_enabled']:
        print 'This host is not ready to perform volume operations, please run "infinihost settings check"'
        raise SystemExit(1)
    elif readiness['reboot_required']:
        print 'NOTICE: Please use "infinihost settings check" to detect changes triggered by the last reboot'


def run_volume_provision(arguments):
    from ..volume_provision.engine import in_provisioning, get_workout
    assert_readiness_for_provisioning()
    prepare_provision_arguments(arguments)
    are_you_sure(arguments, get_workout)
    return in_provisioning(arguments)


def run_volume_resize(arguments):
    from infi.vendata.powertools.volume_resize.engine import in_resize, in_check_preconditions, get_workout
    assert_readiness_for_provisioning()
    res = in_check_preconditions(arguments)
    if res != 0:
        return res
    are_you_sure(arguments, get_workout)
    return in_resize(arguments)


def run_volume_unmap(arguments):
    from infi.vendata.powertools.volume_unmap.engine import in_unmap, get_workout
    assert_readiness_for_provisioning()
    device = get_storagemodel_device(arguments)
    volume_name = device.get_vendor().get_volume_name()
    volume_id = device.get_vendor().get_volume_id()
    message = "Volume \"{}\" (id {}) successfully unmapped.".format(volume_name, volume_id)
    are_you_sure(arguments, get_workout)
    result = in_unmap(device)
    if result == 0:
        print message
    return result


def run_volume_mount(arguments):
    from ..volume_provision.engine import mount, ProvisioningError
    from .validate import validate_mount
    from .autodetect import get_suggested_mount_point
    from infi.vendata.powertools.utils.queries import is_device_read_only
    assert_readiness_for_provisioning()
    device = get_storagemodel_device(arguments)
    if arguments["<mount-path>"] is None:
        arguments["<mount-path>"] = get_suggested_mount_point(device.get_vendor().get_volume_name())
        print "Volume will be mounted as %s" % arguments["<mount-path>"]
    validation_result = validate_mount(arguments["<mount-path>"])
    if len(validation_result) > 0:
        print validation_result
        return 2
    read_only = is_device_read_only(device)
    if read_only and not arguments['--allow-read-only']:
        print "Volume is either a snapshot or write-protected; to mount it in read-only, pass --allow-read-only"
        return 2
    try:
        mount(device, arguments["<mount-path>"], read_only=read_only)
        return 0
    except ProvisioningError as error:
        print error.message
        return 1


def run_volume_rename(arguments):
    from infi.vendata.powertools.volume_rename.engine import rename_device
    from .validate import validate_volume_name, ValidationException
    assert_readiness_for_provisioning()
    device = get_storagemodel_device(arguments)
    validation_result = validate_volume_name(arguments["<name>"], device.get_vendor().get_system_serial())
    if validation_result:
        raise ValidationException(validation_result)
    name = arguments["<name>"].strip()
    result = rename_device(device, name)
    if result == 0:
        print "Volume renamed to %s" % name
    return result


def run_volume_delete(arguments):
    from infi.vendata.powertools.volume_unmap.engine import delete_device
    assert_readiness_for_provisioning()
    device = get_storagemodel_device(arguments)
    are_you_sure(arguments, lambda args: "This command will unmount and delete the selected device.")
    result = delete_device(device)
    if result == 0:
        print "Volume deleted"
    return result


def run_volume_map(arguments):
    from infinisdk.core.exceptions import APICommandFailed
    from ..volume_provision.engine import ProvisioningError
    from ..volume_map.engine import run, get_workout
    autodetect(arguments, "system")
    validate(arguments, "system", only_registered=True)
    try:
        are_you_sure(arguments, get_workout)
        volume_obj = run(arguments)
        print "Volume \"{}\" (id {}) successfully mapped.".format(volume_obj.get_name(), volume_obj.get_id())
    except ProvisioningError as error:
        print error.message
        return 1
    except APICommandFailed as exc:
        print "ERROR: {}".format(exc.message)
        return 1
    if not is_filesystem_support_disabled():
        print "To mount the device, run infinihost volume mount {} [<mount-path>]".format(volume_obj.get_name())
    return 0


def run_volume_create_filesystem(arguments):
    from ..volume_provision.engine import create_filesystem, ProvisioningError
    assert_readiness_for_provisioning()
    device = get_storagemodel_device(arguments)
    autodetect(arguments, "filesystem")
    validate(arguments, "filesystem")
    try:
        create_filesystem(device, arguments["--filesystem"])
        return 0
    except ProvisioningError as error:
        print error.message
        return 1


def run_volume_unmount(arguments):
    from infi.vendata.powertools.volume_unmap.engine import unmount
    assert_readiness_for_provisioning()
    device = get_storagemodel_device(arguments)
    are_you_sure(arguments, lambda args: "This command will unmount the selected device.")
    result = unmount(device)
    if result == 0:
        print "Volume unmounted"
    return result


def run_volume_speedtest(arguments):
    from ..volume_speedtest.engine import PerformanceEngine, SpeedTestException
    assert_readiness_for_provisioning()
    device = get_storagemodel_device(arguments)
    device_path = device.get_block_access_path()
    validate(arguments, "duration")
    time_to_run = int(arguments["--duration"])
    try:
        PerformanceEngine(device_path, time_to_run=time_to_run).run()
    except SpeedTestException as error:
        print error.message
        return 1
    return 0


def run_system_list(arguments):
    from ..system_register.engine import Abort
    from ..system_list.engine import show
    try:
        show()
        return 0
    except Abort:
        return 1


def run_system_register(arguments):
    from infinisdk.core.exceptions import APICommandFailed
    from ..system_register.engine import in_register
    autodetect(arguments, "system", "hostname")
    validate(arguments, "system", "hostname")
    try:
        return in_register(arguments)
    except APICommandFailed as exc:
        print "ERROR: {}".format(exc.message)
        return 1


def run_rescan(arguments):
    from infi.storagemodel.errors import RescanIsNeeded, TimeoutError
    try:
        do_rescan()
        return 0
    except (RescanIsNeeded, TimeoutError):
        raise RescanFailed


def run_logs_collect(arguments):
    from infi.vendata.powertools.logs_collect.engine import run, FreeSpaceError
    try:
        result, archive_path = run(arguments["--timestamp"], arguments["--span"], arguments["--file"],
                                   interactive=arguments['--interactive'],
                                   only_infinihost_logs=arguments['--only-infinihost-logs'])
    except FreeSpaceError:
        # error message already printed
        return 1
    return result


def run_metadata_sync(arguments):
    from ..metadata_sync.engine import metadata_sync
    metadata_sync()
    return 0


def run_credentials_set(arguments):
    from ..utils import auth
    username = (arguments.get("<username>", "") or auth.CREDENTIALS_STORE.ask_for_username()).strip()
    password = (arguments.get("<password>", "") or auth.CREDENTIALS_STORE.ask_for_password()).strip()
    credentials = Credentials(username, password)
    if arguments["--system"] == "autodetect":
        auth.CREDENTIALS_STORE.set_default_credentials(credentials)
    else:
        validate(arguments, "system")
        auth.CREDENTIALS_STORE.set_credentials(arguments['--system'], credentials)
    print "Credentials for user %s were saved" % username


def run_credentials_get(arguments):
    from ..utils.tables import Table, Column
    from base64 import b64encode
    from ..utils import auth
    table = Table([Column(0, "InfiniBox Serial"), Column(1, "Username")])
    _dict = auth.CREDENTIALS_STORE.get_credentials_dict()
    if _dict:
        for serial, credentials in _dict.items():
            table.add_row([serial, credentials.get_username(), b64encode(credentials.get_password())])
        print table
    else:
        print "No saved credentials found"


def run_credentials_delete(arguments):
    from ..utils import auth
    auth.CREDENTIALS_STORE.reset_credentials()
    print "Saved credentials deleted"


def ensure_root():
    from . import get_privileges_text
    from infi.run_as_root import RootPermissions

    if not RootPermissions().is_root():
        raise docopt.DocoptExit(get_privileges_text())


@suppress_warnings
@debug_program()
@consider_mock_environment
@log_usage
def run_command(arguments):
    from infi.vendata.powertools.cli.os_check import check_os_support
    check_os_support()
    init_colors()
    ensure_root()

    CMD_MAP = {'settings':      {'check': run_settings_check},
               'connectivity':  {'list': run_connectivity_list},
               'pool':          {'list': run_pool_list},
               'volume':        {'list': run_volume_list,
                                 'provision': run_volume_provision,
                                 'resize': run_volume_resize,
                                 'unmap': run_volume_unmap,
                                 'speedtest': run_volume_speedtest,
                                 'unmount': run_volume_unmount,
                                 'create-filesystem': run_volume_create_filesystem,
                                 'mount': run_volume_mount,
                                 'delete': run_volume_delete,
                                 'map': run_volume_map,
                                 'rename': run_volume_rename},
               'system':        {'list': run_system_list,
                                 'register': run_system_register},
               'rescan':        run_rescan,
               'logs':          {'collect': run_logs_collect},
               'metadata':      {'sync': run_metadata_sync},
               'credentials':   {'set': run_credentials_set,
                                 'get': run_credentials_get,
                                 'delete': run_credentials_delete}
               }

    cur_cmd_map = CMD_MAP
    while not callable(cur_cmd_map):
        for key, cmd_map in cur_cmd_map.items():
            if arguments.get(key, False):
                cur_cmd_map = cmd_map
                break
        else:
            # this shouldn't really happen if docopt works
            raise docopt.DocoptExit("No command found.")

    cmd_entry = cur_cmd_map
    return cmd_entry(arguments)
