import re
from os import mkdir, rmdir


class ValidationException(Exception):
    pass


def get_storagemodel_device(arguments):
    """ find storagemodel device according to <device_or_mount> """
    from infi.storagemodel.base.multipath import MultipathBlockDevice
    from infi.storagemodel.base.disk import NoSuchDisk
    from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices
    from infi.vendata.powertools.cli.engine import is_filesystem_support_disabled
    from ..utils import get_mount_point
    from os.path import normpath, normcase
    for device in get_infinidat_block_devices():
        # Check device name / path / id
        try:
            block_access_path = device.get_disk_drive().get_block_access_path()
        except (NoSuchDisk, NotImplementedError):
            block_access_path = None
        possible_devices = [device.get_vendor().get_volume_name(),
                            device.get_display_name(),
                            device.get_block_access_path(),
                            block_access_path]
        if isinstance(device, MultipathBlockDevice):
            possible_devices.extend([str(path.get_path_id()) for path in device.get_paths()])
        if arguments["<device-or-mount>"] in possible_devices or arguments.get("<device>", "") in possible_devices:
            return device
        # Check device mount point
        if not is_filesystem_support_disabled():
            if arguments["<device-or-mount>"] is not None:
                mount_point = get_mount_point(device)
                if mount_point and normpath(normcase(mount_point)) == normpath(normcase(arguments["<device-or-mount>"])):
                    return device
    raise ValidationException("No valid device/mountpoint was found")

def convert_size(string):
    """this function parses the size string
    :returns: a infi.dtypes object
    :raises: argparse.ArgumentTypeError
    """
    from capacity import from_string, byte
    from ..utils import round_to_block_size
    try:
        return round_to_block_size(from_string(string) // byte) * byte
    except (ValueError, TypeError):
        pass
    try:
        return round_to_block_size(from_string(string.upper()) // byte) * byte
    except (ValueError, TypeError):
        raise ValidationException("Invalid size")

def validate_name(string, objtype='volume'):
    if not string.strip():
        return "{} name cannot be empty".format(objtype)
    if re.match('^[a-zA-Z0-9\\+\\^&\'@{}\\$=!()#%.~_ -]{1,65}$', string) is None:
        return "{} is not a valid {} name".format(string, objtype)
    return ""

def validate_volume_name(volume_name, system_serial):
    from .autodetect import volume_exists
    validation_result = validate_name(volume_name)
    if validation_result:
        return validation_result
    if volume_exists(system_serial, volume_name):
        return "Volume name already exists"

def _validate_name(arguments):
    string = arguments.get('<name>') or arguments.get("--name")
    validation_result = validate_volume_name(string, arguments["--system"])
    if validation_result:
        raise ValidationException(validation_result)

def is_writable(dirpath):
    try:
        mkdir(dirpath)
    except OSError:
        return False
    rmdir(dirpath)
    return True

def validate_mount(string):
    from os.path import exists, pardir, join, abspath, ismount, isdir
    from os import name, listdir
    import re
    if not string:
        return "mountpoint cannot be an empty string"
    parent_directory = abspath(join(string, pardir))
    current_directory = abspath(string)
    if '\x00' in string:  # \x00 is the only invalid character in POSIX
        return u"{!r} is an invalid mountpoint".format(string)
    if name != "nt" and not exists(parent_directory):
        return u"{} does not exist".format(parent_directory)
    if exists(current_directory) and ismount(current_directory):
        return u"{} is already mounted".format(current_directory)
    if exists(current_directory) and not isdir(current_directory):
        return u"{} exists but its not a directory".format(current_directory)
    if exists(current_directory) and listdir(current_directory):
        return u"{} is not empty".format(current_directory)
    if name != "nt" and exists(parent_directory) and not exists(current_directory) and not is_writable(current_directory):
        return u"{} is not writable".format(parent_directory)
    if name == "nt":
        if re.match("^[A-Za-z]:\\\?$", string) is None:
            # its not just a drive letter
            if re.match("^[A-Za-z]:\\\[A-Za-z0-9_.,\\-\\\\]+$", string) is None:
                return u"{} is not a valid mount point".format(string)
            if not exists(parent_directory):
                return u"{} does not exist".format(parent_directory)
    return ""

def validate_filesystem_size(filesystem, size):
    from capacity import TB
    filesystem_limit = 16*TB       # HIP-727 we decided not to let filesystems be too large regardless of type
    if size > filesystem_limit:
        msg = "Size exceeds maximum allowed filesystem size"
        return msg
    return ""

def _validate_filesystem_size(filesystem, size):
    validation_result = validate_filesystem_size(filesystem, size)
    if validation_result:
        raise ValidationException(validation_result)

def validate_pool_size(system_serial, pool_name, size, count=1):
    from infi.vendata.powertools.utils.queries import get_system_object_by_serial
    from capacity import byte
    system = get_system_object_by_serial(system_serial)
    pool = system.pools.get(name=pool_name)
    if size * count > pool.get_free_virtual_capacity():
        return "Not enough free space in pool"
    # also validate minimum volume size
    minimum_volume_size = byte * system.api.get("config/limits/volumes.minimum_size").get_result()
    if size < minimum_volume_size:
        return "Volume size is lower than the minimum volume size allowed."
    return ""

def _validate_pool_size(arguments):
    # assuming system, pool and size are already validated and size is already converted
    validation_result = validate_pool_size(arguments["--system"], arguments["--pool"], arguments["<size>"])
    if validation_result:
        raise ValidationException(validation_result)

def _validate_mount(arguments):
    string = arguments["--mount"]
    validation_result = validate_mount(string)
    if validation_result:
        raise ValidationException(validation_result)

def _validate_with_options(arguments, key, options):
    if arguments["--" + key] not in options:
        raise ValidationException("Invalid option for {}. Choices are: {}".format(key, ', '.join(options)))

def _validate_filesystem(arguments):
    from autodetect import get_list_of_supported_filesystems
    options = get_list_of_supported_filesystems()
    _validate_with_options(arguments, "filesystem", options)

def _validate_type(arguments):
    options = ["thick", "thin"]
    _validate_with_options(arguments, "type", options)

def _validate_thin(arguments):
    options = ["yes", "no"]
    _validate_with_options(arguments, "thin", options)

def _validate_pool(arguments):
    from autodetect import get_available_pools
    options = get_available_pools(arguments["--system"])
    _validate_with_options(arguments, "pool", options)

def _system_name_to_serial(arguments, system_options):
    """ convert the system argument to serial, if and only if it is currently a system name """
    name_options = [serial for serial, name in system_options if arguments["--system"] == name]
    if len(name_options) == 1:
        arguments["--system"] = str(name_options[0])

def _validate_system(arguments, only_registered=False):
    from autodetect import get_system_serials_of_attached_boxes
    options = get_system_serials_of_attached_boxes(False)
    _system_name_to_serial(arguments, options)
    chosen_serial = arguments["--system"]
    serials = [str(serial) for serial, name in options]
    if not serials:
        raise ValidationException("Host is not connected to any system")
    if chosen_serial not in serials:
        options = ", ".join([", ".join([str(serial), name]) for serial, name in options])
        raise ValidationException("Invalid option for system. Choices are: {}".format(options))
    if only_registered:
        registered_systems = get_system_serials_of_attached_boxes(True)
        serials = [str(serial) for serial, name in registered_systems]
        if chosen_serial not in serials:
            raise ValidationException("Host is not registered in system {0}, please run 'infinihost system register --system={0} --fix'".format(chosen_serial))

def _validate_lun(arguments):
    if arguments.get('--lun') is None:
        return
    if not arguments['--lun'].isdigit():
        raise ValidationException("LUN number needs to be an a non-negative integer")
    arguments['--lun'] = int(arguments['--lun'])
    if arguments['--lun'] < 0 or arguments['--lun'] > 127:
        raise ValidationException("LUN number needs to at least 0 and no more than 127")

def _validate_hostname(arguments):
    string = arguments["--hostname"]
    validation_result = validate_name(string, objtype='host')
    if validation_result:
        raise ValidationException(validation_result)

def _validate_duration(arguments):
    if not arguments["--duration"].isdigit():
        raise ValidationException("duration must be a an integer (seconds)")

def validate(arguments, *keys, **kwargs):
    for key in keys:
        validator = globals().get("_validate_" + key)
        validator(arguments, **kwargs)
