from .container import Container
from infi.vendata.powertools.utils.file import gevent_async_open, safe_write_file
from pkg_resources import resource_filename
import json
import os


DEFAULTS_FILEPATH = resource_filename(__name__, "defaults.json")

class Configuration(Container):
    def __init__(self):
        defaults = self._load_default_config()
        self._ensure_default_config_exists()
        custom = self._load_custom_config()
        running = update_dict_recursively(defaults, custom)
        super(Configuration, self).__init__(running)

    def _load_default_config(self):
        return read_json_from_file(DEFAULTS_FILEPATH)

    def _load_custom_config(self):
        self._ensure_custom_config_exists()
        return read_json_from_file(self._get_custom_config_path())

    def _safe_write_file(self, filepath, content):
        with safe_write_file(filepath) as fd:
            fd.write(content)

    def _ensure_default_config_exists(self):
        path = self._get_default_config_path()
        with gevent_async_open(DEFAULTS_FILEPATH, 'r') as src:
            self._safe_write_file(path, src.read())

    def _ensure_custom_config_exists(self):
        path = self._get_custom_config_path()
        if not os.path.exists(path):
            self._safe_write_file(path, '{}')

    def _get_custom_config_path(self):
        return os.path.join(self._get_root_path(), 'conf', 'custom.json')

    def _get_default_config_path(self):
        return os.path.join(self._get_root_path(), 'conf', 'defaults.json')

    def _get_root_path(self):
        # We put the conf directory under PROJECTROOT, but if powertools is
        # installed as an egg we need to move up to the parent directory
        from infi.vendata.powertools import PROJECTROOT
        return os.path.dirname(PROJECTROOT) if PROJECTROOT.endswith('eggs') else PROJECTROOT


def read_json_from_file(src):
    with gevent_async_open(src, 'r') as fd:
        return json.load(fd)


def update_dict_recursively(d, u):
    # http://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth
    from collections import Mapping
    for k, v in u.iteritems():
        if isinstance(v, Mapping):
            r = update_dict_recursively(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d


_configuration = Configuration()

def get_configuration():
    return _configuration
