from __future__ import print_function
from ...utils.logging import debug_program, suppress_warnings
from ...utils.mock_environment import consider_mock_environment
from infi.vendata.powertools.__version__ import __version__

def parse_args():
    import argparse
    parser = argparse.ArgumentParser(description="INFINIDAT Host PowerTools GUI",
                                 version=__version__)
    parser.add_argument("--no-window", default=False, action="store_true", required=False,
                        help="If present, the GUI server will open without the window")
    parser.add_argument("--browser", default=False, action="store_true", required=False,
                        help="If present, the default browser will open with the GUI page")
    parser.add_argument("--port", default=0, action="store", type=int, required=False,
                        help="Network port on which to open the GUI page")
    namespace = parser.parse_args()
    return namespace

def with_gevent_patching(func):
    from infi.pyutils.decorators import wraps
    @wraps(func)
    def decorated_function(*args, **kwargs):
        from gevent.monkey import patch_all
        patch_all()
        return func(*args, **kwargs)
    return decorated_function

@suppress_warnings
@debug_program()
@with_gevent_patching
@consider_mock_environment
def development_server():  # pragma: no cover
    from infi.vendata.powertools.utils.progress_bar import disable
    from ..webserver import WebServer
    from .starter import _start_all
    import warnings
    if not check_os_support():
        return 1
    namespace = parse_args()
    with warnings.catch_warnings():
        import infinisdk  # noqa
    if not _start_all(namespace):
        return
    disable()
    try:
        WebServer.serve_forever()
    except KeyboardInterrupt:
        # Developer stopped the server
        return
    finally:
        WebServer.stop()


def check_os_support():
    from infi.os_info import get_platform_string
    from infi.vendata.powertools.utils.mock_environment import SMOCK_KEY
    from os import environ
    use_mock = bool(environ.get(SMOCK_KEY, False))
    platform_tuple = get_platform_string().split("-")
    if use_mock:  # the mock fakes the platform as aix-7.1-powerpc
        return True
    os = platform_tuple[0]
    if os == "solaris":
        _print_warning("infinihost-gui is not supported on Solaris")
        return False
    elif os == "aix":
        _print_warning("infinihost-gui is not supported on AIX")
        return False
    return True

def _print_warning(msg):
    from infi.vendata.powertools.settings.runner import get_fore_color
    from sys import stderr
    print(get_fore_color("RED") + "ERROR: " + get_fore_color("RESET") + msg, file=stderr)
