from logging import getLogger
import os

logger = getLogger(__name__)

def _disable_caching_for_staticdir():
    from infi.pyutils.decorators import wraps
    import cherrypy
    import mock
    patcher = mock.patch.object(cherrypy.lib.static.cptools, 'validate_since')
    def patch_calls_to_validate_since(func):
        @wraps(func)
        def callable(*args, **kwargs):
            with patcher as patch:
                patch.return_value = False
                return func(*args, **kwargs)
        return callable
    cherrypy.lib.static.staticdir = patch_calls_to_validate_since(cherrypy.lib.static.staticdir)


def get_gui_url():
    from ...webserver import WebServer
    from ...webapp.plugin.resource import AuthKey
    address, port = WebServer.get_address_and_port()
    authkey = AuthKey.get_authkey()
    return "http://{}:{}/plugin/views/host/?auth_key={}".format(address, port, authkey)


def open_browser():
    import webbrowser
    try:
        default = webbrowser.get()  # when no browser is found
    except webbrowser.Error:
        return
    if isinstance(default, webbrowser.GenericBrowser):  # this happens on OSX and over SSH
        if default.basename == 'www-browser':  # over SSH
            pass
        if default.basename in ('links', 'elinks', 'lynx', 'w3m'):  # browsers we do not support
            pass
        elif default.basename == 'open':  # OSX-specific
            webbrowser.open(get_gui_url())
    else:
        webbrowser.open(get_gui_url())


def open_node_webkit():
    from infi.node_webkit import get_node_webkit_executable
    from infi.execute import execute_async
    import gevent
    import sys
    # Run node-webkit and exit when the user closes it
    def run_and_wait():
        try:
            config_dir = os.path.dirname(__file__)
            cmd = [get_node_webkit_executable(), config_dir, get_gui_url()]
            logger.info(' '.join(cmd))
            nw = execute_async(cmd)
        except:
            logger.exception('Cannot run node-webkit')
            print 'Cannot open native GUI window, please visit the URL above in your browser instead.'
            return
        nw.wait()
        logger.info('node-webkit output:\n%s\n%s', nw.get_stdout(), nw.get_stderr())
        if nw.get_returncode() == 127:  # HIP-1443
            # .../nw: error while loading shared libraries:
            # libgio-2.0.so.0: cannot open shared object file: No such file or directory
            return
        sys.exit(0)
    gevent.spawn(run_and_wait)


def _patch_multiprocessing():
    # on Windows, infinihost-gui is a gui script (no console), but the volume speedtest uses multiprocessing
    # which causes a console to open for the subprocesses. We patch CreateProcess to pass the NO_WINDOW flag
    if os.name != 'nt':
        return
    from _subprocess import CreateProcess
    import _subprocess
    from mock import patch
    def PatchedCreateProcess(app_name, cmd_line, proc_attrs, thread_attrs,
                             inherit, flags, env_mapping, curdir, startup_info):
        CREATE_NO_WINDOW = 0x08000000
        flags |= CREATE_NO_WINDOW
        return CreateProcess(app_name, cmd_line, proc_attrs, thread_attrs,
                             inherit, flags, env_mapping, curdir, startup_info)
    create_process_patch = patch.object(_subprocess, "CreateProcess", new=PatchedCreateProcess)
    create_process_patch.start()


def _start_all(namespace):  # pragma: no cover
    from ...webserver import WebServer
    from infi.run_as_root import ensure_root
    import socket
    ensure_root()
    _disable_caching_for_staticdir()
    _patch_multiprocessing()
    try:
        WebServer.start(namespace.port)
    except socket.error as e:
        if e.errno not in [48, 10048, 98]:      # address already in use errors - Mac, Windows, Linux
            raise
        # address already in use
        print "GUI server is already running."
        if os.name == "nt":
            import ctypes
            ctypes.windll.user32.MessageBoxA(None, "GUI Server is already running.", "Error", 0)
        return False
    print "GUI page is available in the following URL:\n", get_gui_url()
    if not namespace.no_window:
        open_node_webkit()
    if namespace.browser:
        open_browser()
    return True
