import cherrypy
import os
import json
from infi.pyutils.decorators import wraps

STATIC_PATH = os.path.join(os.path.dirname(__file__), 'static')
TEMPLATES_PATH = os.path.join(os.path.dirname(__file__), 'templates')
FAVICON_PATH = os.path.join(STATIC_PATH, 'images', 'icon-32.png')

from logging import getLogger
logger = getLogger(__name__)

def is_unexpected_error(error):
    # The client shows "Unexpected error has occurred" when there are exceptions,
    # but some exceptions are not as unexpected (e.g. "Volume name already exists") so we can
    # show the message from the error instead of showing "Unexpected ..."
    # here is a list of exceptions that are "expected", for which the message is informative
    from infinisdk.core.exceptions import APICommandFailed, APITransportFailure, SystemNotFoundException
    from infi.vendata.powertools.volume_provision.engine import ProvisioningError
    from infi.vendata.powertools.volume_resize.engine import ResizeError
    from infi.vendata.powertools.cli.validate import ValidationException
    from infi.vendata.powertools.gui.webapp.plugin.views.host.operations import ProvisionValidationException
    from infi.vendata.powertools.utils.queries import ObjectNotFound
    return not isinstance(error, (ValidationException, ProvisioningError, APICommandFailed, SystemNotFoundException,
                                  APITransportFailure, ProvisionValidationException, ResizeError, ObjectNotFound))

def json_response(func):
    @wraps(func)
    def callable(*args, **kwargs):
        if cherrypy.request.method in ('PUT', 'POST'):
            data = json.loads(cherrypy.request.body.read() or '{}')
            if isinstance(data, dict):
                kwargs.update(data)
            else:
                kwargs['data'] = data
        cherrypy.response.headers['Content-Type'] = "application/json"
        error_message = None
        return_value = None
        unexpected_error = False
        try:
            return_value = func(*args, **kwargs)
            success = True
        except (Exception, SystemExit) as error:
            from infinisdk.core.exceptions import APICommandFailed, APITransportFailure, SystemNotFoundException
            if isinstance(error, APICommandFailed):
                error_message = error.message
            elif isinstance(error, (SystemNotFoundException, APITransportFailure)):
                msg = "Failed to communicate with system. " \
                      "Please check IP connectivity to \"{}\" from the PowerTools server"
                error_message = msg.format(error.address)
            else:
                error_message = unicode(error)
            if len(error_message.strip()) == 0:
                error_message = error.__class__.__name__
            unexpected_error = is_unexpected_error(error)
            msg = "Caught an exception on json response on function call {} with args {!r}, kwargs {!r}"
            logger.exception(msg.format(func.__name__, args, kwargs))
            success = False
        return json.dumps(dict(success=success, return_value=return_value, error_message=error_message,
            unexpected_error=unexpected_error))
    callable.exposed = True
    return callable

from .plugin import Plugin

class WebApplication(object):
    def __init__(self):
        super(WebApplication, self).__init__()

    @cherrypy.expose
    def index(self):
        pass

    @cherrypy.expose
    def relaunch(self):
        from ..scripts.starter import open_browser
        open_browser()

    @cherrypy.expose
    def shutdown(self):
        from ..webserver import WebServer
        WebServer.stop()

    @cherrypy.expose
    def log(self, **kwargs):
        from ...utils.logging import DebugLogging
        cherrypy.response.headers['Content-Type'] = "text/plain"
        log_path = DebugLogging().get().get_logfile_path()
        return open(log_path, "rb").read()

    plugin = Plugin()

    @classmethod
    def get_wsgi_application(cls, authkey_verification=True):
        from .error_pages import error_page
        from .plugin.resource import set_request_hook
        set_request_hook()
        logging_configuration = {'log.access_file': '',
                                 'log.error_file': '',
                                 'log.screen': False,
                                 'error_page.401': os.path.join(STATIC_PATH, 'html', '401.html'),
                                 'error_page.404': os.path.join(STATIC_PATH, 'html', '404.html'),
                                 'error_page.default': error_page,
                                }
        cherrypy.config.update(logging_configuration)
        configuration = {'/plugin/views': {'tools.assert_browser.on': True,
                                           'tools.assert_authkey.on': authkey_verification},
                         '/favicon.ico': {'tools.staticfile.on': True,
                                          'tools.staticfile.filename': FAVICON_PATH},
                         '/static': {'tools.staticdir.on': True,
                                     'tools.staticdir.dir': STATIC_PATH},
                         '/log': {'tools.assert_authkey.on': authkey_verification},
                        }

        app = cherrypy.tree.mount(cls(), '/', configuration)
        return app
