import logging
from ......webapp import json_response
from infi.pyutils.lazy import cached_function, clear_cache
from .resize import Resize

logger = logging.getLogger(__name__)

class ProvisionValidationException(Exception):
    pass

def map_existing_volume(system_serial, volume_id):
    from infi.vendata.powertools.volume_map.engine import map_existing_volume
    system_serial = int(system_serial)
    volume_id = int(volume_id)
    map_existing_volume(system_serial, volume_id)

def get_provision_params(box_serial):
    from infi.vendata.powertools.cli.autodetect import get_list_of_supported_filesystems
    from infi.vendata.powertools.cli.autodetect import get_suggested_mount_point
    from infi.vendata.powertools.cli.autodetect import get_suggested_volume_name
    from infi.vendata.powertools.cli.autodetect import get_default_filesystem_choice
    from infi.vendata.powertools.utils.queries import get_system_object_by_serial, get_pools
    import capacity
    filesystem_choices = [dict(name=name) for name in get_list_of_supported_filesystems()]
    filesystem = get_default_filesystem_choice(filesystem_choices)
    volume_name = get_suggested_volume_name(box_serial)
    mountpoint = get_suggested_mount_point(volume_name)
    system = get_system_object_by_serial(box_serial)
    pools = [dict(name=pool.get_name(),
                  size=pool.get_virtual_capacity() // capacity.byte,
                  free=pool.get_free_virtual_capacity() // capacity.byte) for pool in get_pools(system)]
    pools.sort(key=lambda pool: pool["name"].lower())
    return dict(box_serial=box_serial,
                filesystems=filesystem_choices,
                selected_filesystem=filesystem,
                suggested_volume_name=volume_name,
                suggested_mountpoint=mountpoint,
                pools=pools)

def validate_size(volume_size):
    if not volume_size.endswith("TB") and not volume_size.endswith("GB"):
        raise ProvisionValidationException("No volume size unit found")
    try:
        float(volume_size[:-2])
    except ValueError:
        raise ProvisionValidationException("Volume size is invalid")

def provision_validation(box_serial, filesystem, volume_name, mountpoint, pool_name, volume_size, thin, do_filesystem, do_mount, lun):
    from infi.vendata.powertools.cli.autodetect import get_list_of_supported_filesystems
    from infi.vendata.powertools.cli.autodetect import get_available_pools
    from infi.vendata.powertools.cli.validate import validate_volume_name, validate_mount
    from infi.vendata.powertools.cli.validate import convert_size, validate_filesystem_size
    from infi.vendata.powertools.cli.validate import validate_pool_size
    try:
        int(box_serial)
    except ValueError:
        raise ProvisionValidationException("Box serial is invalid (expected a number)")
    if lun is not None:
        try:
            int(lun)
        except ValueError:
            raise ProvisionValidationException("LUN is invalid (expected a number)")
    validate_size(volume_size)
    if filesystem and filesystem not in get_list_of_supported_filesystems():
        raise ProvisionValidationException("Unavailable filesystem")
    if pool_name not in get_available_pools(box_serial):
        raise ProvisionValidationException("Unavailable pool")
    validation_result = validate_volume_name(volume_name, box_serial)
    if validation_result:
        raise ProvisionValidationException(validation_result)
    volume_size_converted = convert_size(volume_size)
    validation_result = validate_pool_size(box_serial, pool_name, volume_size_converted)
    if validation_result:
        raise ProvisionValidationException(validation_result)
    if do_filesystem:
        validation_result = validate_filesystem_size(filesystem, volume_size_converted)
        if validation_result:
            raise ProvisionValidationException(validation_result)
        if do_mount:
            validation_result = validate_mount(mountpoint)
            if validation_result:
                raise ProvisionValidationException(validation_result)

def prepare_provision_args(box_serial, filesystem, volume_name, mountpoint, pool_name, volume_size, thin, do_filesystem, do_mount, lun):
    import os
    from munch import Munch
    from infi.vendata.powertools.cli.validate import convert_size
    provision_validation(box_serial, filesystem, volume_name, mountpoint, pool_name, volume_size, thin, do_filesystem, do_mount, lun)
    volume_size = convert_size(volume_size)
    args = Munch()
    args.box_id = int(box_serial)
    args.filesystem = filesystem
    args.volume_name = volume_name
    args.mountpoint = mountpoint
    args.pool_name = pool_name
    args.volume_size = volume_size
    args.volume_type = "thin" if thin else "thick"
    args.do_filesystem = do_filesystem
    args.do_mount = do_mount
    args.are_you_sure = True
    args.lun = None if lun is None else int(lun)
    args.label = volume_name[:32 if os.name == 'nt' else 16]
    return args

@cached_function
def get_provision_tasks_object(box_serial, filesystem, volume_name, mountpoint, pool_name, volume_size, thin, do_filesystem, do_mount, lun):
    from infi.vendata.powertools.volume_provision.engine import load_tasks
    args = prepare_provision_args(box_serial, filesystem, volume_name, mountpoint,
                                  pool_name, volume_size, thin, do_filesystem, do_mount, lun)
    return load_tasks(args)

def get_provision_tasks(box_serial, filesystem, volume_name, mountpoint, pool_name, volume_size, thin, do_filesystem, do_mount, lun):
    logger.debug("Getting provision task list")
    # we want to cache the provision volume from now and throguh the actions, but earlier caches are invalid
    clear_cache(get_provision_tasks_object)
    task_sequence = get_provision_tasks_object(box_serial, filesystem, volume_name, mountpoint,
                                               pool_name, volume_size, thin, do_filesystem, do_mount, lun)
    return task_sequence.get_description()

def run_provision_task(action_description, box_serial, filesystem, volume_name, mountpoint, pool_name, volume_size, thin, do_filesystem, do_mount, lun):
    from infi.storagemodel import get_storage_model
    task_sequence = get_provision_tasks_object(box_serial, filesystem, volume_name, mountpoint,
                                               pool_name, volume_size, thin, do_filesystem, do_mount, lun)
    completed_tasks = []
    for task in task_sequence.tasks:
        if task.get_description() == action_description:
            logger.debug("Running provision task {}".format(action_description))
            try:
                task.run()
                # some operations run in a subprocess, so rescan doesn't reflect in the current process - so clear cache
                get_storage_model().refresh()
            except BaseException as ex:
                logger.exception("exception during provisioning")
                # we don't return "failed" or some string because we want to show an error popup instead of marking
                # the task as failed in the operation table. So we delegate this to @json_response and view.js:ajax
                # will take care of the rest.
                raise ex
            logger.debug("Provision task succeeded")
            return "ok"
        # we run the tasks in order so if we got here it means the task was run in a previous call.
        # the task will be the same instance that we ran because the list is cached
        completed_tasks.append(task)

def get_storagemodel_device_by_serial(serial):
    from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices
    serial = str(serial.replace(u'\ufffd', '\x00'))
    device = [device for device in get_infinidat_block_devices()
              if device.get_scsi_serial_number() == serial]
    if device == []:
        message = "Device {} does not exist on host".format(serial)
        logger.info(message)
        return None
    device = device[0]
    return device

class Operations(object):
    @json_response
    def map_existing_volume(self, serial, volume_id, **kwargs):
        map_existing_volume(serial, volume_id)

    @json_response
    def get_provision_params(self, serial, **kwargs):
        return get_provision_params(serial)

    @json_response
    def validate_provision_params(self, serial, filesystem, volume_name, mountpoint, pool_name, volume_size, thin, do_filesystem, do_mount, lun, **kwargs):
        return provision_validation(serial, filesystem, volume_name, mountpoint, pool_name, volume_size, thin, do_filesystem, do_mount, lun)

    @json_response
    def run_provision_task(self, action_description, serial, filesystem, volume_name, mountpoint, pool_name, volume_size, thin, do_filesystem, do_mount, lun, **kwargs):
        return run_provision_task(action_description, serial, filesystem, volume_name, mountpoint, pool_name, volume_size, thin, do_filesystem, do_mount, lun)

    @json_response
    def get_provision_tasks(self, serial, filesystem, volume_name, mountpoint, pool_name, volume_size, thin, do_filesystem, do_mount, lun, **kwargs):
        return get_provision_tasks(serial, filesystem, volume_name, mountpoint, pool_name, volume_size, thin, do_filesystem, do_mount, lun)

    @json_response
    def create_fs(self, serial, filesystem_type, **kwargs):
        from infi.vendata.powertools.volume_provision.engine import create_filesystem
        create_filesystem(get_storagemodel_device_by_serial(serial), filesystem_type, True)

    @json_response
    def is_read_only(self, serial, **kwargs):
        from infi.vendata.powertools.utils.queries import is_device_read_only
        device = get_storagemodel_device_by_serial(serial)
        return dict(read_only=is_device_read_only(device))

    @json_response
    def mount(self, serial, path, **kwargs):
        from infi.vendata.powertools.volume_provision.engine import mount
        from infi.vendata.powertools.cli.validate import validate_mount
        from infi.vendata.powertools.utils.queries import is_device_read_only
        validation_result = validate_mount(path)
        if len(validation_result) > 0:
            raise ProvisionValidationException(validation_result)
        device = get_storagemodel_device_by_serial(serial)
        read_only = is_device_read_only(device)
        mount(device, path, read_only=read_only, silent=True)

    @json_response
    def unmount(self, serial, **kwargs):
        from infi.vendata.powertools.volume_unmap.engine import unmount
        unmount(get_storagemodel_device_by_serial(serial))

    @json_response
    def delete(self, serial, **kwargs):
        from infi.vendata.powertools.volume_unmap.engine import delete_device
        delete_device(get_storagemodel_device_by_serial(serial))

    @json_response
    def unmap(self, serial, **kwargs):
        from infi.vendata.powertools.volume_unmap.engine import unmap
        unmap(get_storagemodel_device_by_serial(serial))

    @json_response
    def rename(self, serial, name, **kwargs):
        from infi.vendata.powertools.volume_rename.engine import rename_device_impl
        rename_device_impl(get_storagemodel_device_by_serial(serial), name, disable_print=True)

    @json_response
    def open_mount_point(self, serial, **kwargs):
        device = get_storagemodel_device_by_serial(serial)
        from infi.vendata.powertools.volume_list.engine import StorageModelDevice
        mount_point = StorageModelDevice(device).get_mount_point()
        from infi.vendata.powertools.utils import open_folder
        open_folder(mount_point)

    @json_response
    def query_mount_after_map(self, serial, volume_id, **kwargs):
        from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_native_multipath_block_devices
        from infi.vendata.powertools.utils import has_filesystem
        # input: system serial, volume id
        # output: whether matching device have a filesystem, and its scsi serial
        # used after mapping a volume, to pop up the "mount volume" dialog if relevant
        serial = int(serial)
        volume_id = int(volume_id)
        matching_devices = [device for device in get_infinidat_native_multipath_block_devices()
                            if volume_id == device.get_vendor().get_naa().get_volume_id() and
                            serial == device.get_vendor().get_naa().get_system_serial()]
        if len(matching_devices) != 1 or not has_filesystem(matching_devices[0]):
            return dict(has_fs=False)
        return dict(has_fs=True, serial=matching_devices[0].get_scsi_serial_number())

    @json_response
    def update_metadata(self, **kwargs):
        from infi.vendata.powertools.metadata_sync.engine import metadata_sync
        metadata_sync(silent=True)

    @json_response
    def rescan(self, **kwargs):
        from infi.vendata.powertools.rescan.engine import do_rescan
        do_rescan(disable_print=True)

    resize = Resize()
