import cherrypy
import logging
from ......webapp import json_response
from infi.vendata.powertools.utils.qrcode import QRCodePool
from infi.vendata.powertools.settings.runner import ReadinessRunner
from infi.vendata.powertools.settings.runner import KB_PREFIX, Status
from infi.pyutils.lazy import cached_method, clear_cache

logger = logging.getLogger(__name__)

class Readiness(object):
    @json_response
    def index(self, **kwargs):
        return getattr(self, cherrypy.request.method.upper())(**kwargs)

    @cached_method
    def _get_runner(self):
        """ Get the ReadinessRunner instance. This is cached to preserve state between POSTs """
        return ReadinessRunner(False, True, False, True)

    def _get_runner_case(self, action_description):
        runner = self._get_runner()
        for case in runner.cases:
            if str(case) == action_description:
                return runner, case
        raise Exception("Readiness case not found")

    def _run_case(self, action_description):
        runner, case = self._get_runner_case(action_description)
        logger.debug("Running readiness case: {}".format(action_description))
        result = runner.run_case(case)
        return case, result

    def POST(self, action_description, **kwargs):
        """ Run one readiness case """
        case, result = self._run_case(action_description)
        msg, msg_color = Status.get_message(result)
        logger.debug("Readiness case status: {}".format(msg))
        # If it's the last case, save the results to file
        runner = self._get_runner()
        if len(runner.cases) == len(runner.result.results):
            runner.save_result()
        return msg

    def GET(self, **kwargs):
        """ Get the host's readiness status from the previous run """
        readiness = ReadinessRunner.load_previous_result()
        if readiness is None:
            readiness = dict(host_ready=False, reboot_required=False, provisioning_enabled=False)
        return readiness

    @json_response
    def list(self, **kwargs):
        """ Called to get the list of cases. We clear the cache here to get a fresh ReadinessRunner instance """
        logger.debug("Getting readiness case list")
        clear_cache(self)
        runner = self._get_runner()
        return [str(case) for case in runner.cases]

    def _get_case_result(self, action_description):
        runner = self._get_runner()
        for case, exception in runner.result.results:
            if str(case) == action_description:
                return case, exception
        raise Exception("Readiness case not found")

    @json_response
    def case_info(self, action_description, **kwargs):
        """ Get additional information about a case. We assume that the case has been run already """
        case, result = self._get_case_result(action_description)
        kb_url = "{}/{}".format(KB_PREFIX, case.get_kb_number())
        info_str = 'For more information, see <a target="_blank" href="{}">{}</a>'.format(kb_url, kb_url)
        qr_img_url = cherrypy.request.path_info.replace("case_info", "qr_img")
        full_qr_img_url = "{}?kb_url={}&{}".format(qr_img_url, kb_url, cherrypy.request.query_string)
        return_dict = {"result": str(result), "info_str": info_str, "kb_img_url": full_qr_img_url}
        return return_dict

    @cherrypy.expose
    def qr_img(self, kb_url, **kwargs):
        """ Create a QR code for the given KB URL """
        cherrypy.response.headers["Content-Type"] = "image/png"
        pool = QRCodePool()
        return pool.get_image_data(kb_url)
