import logging
from ......webapp import json_response
from .performance import Performance
from urllib import urlencode
from infi.vendata.powertools.utils import is_veritas_multipathing_installed

logger = logging.getLogger(__name__)


class HostVolumesTable(object):
    COLUMN_NAMES = ["Volume Name", "Device Name", "Mount", "Size", "# Paths", "Lun", "Actions"]

    def is_device_belongs_to_this_system(self, device, system_serial):
        return device.get_vendor().get_system_serial() == system_serial

    def _get_volume_element(self, infinidat_device):
        vendor = infinidat_device.get_vendor()
        base_url = "https://{}".format(vendor.get_management_address())
        url = "{}/#/volumes/{}".format(base_url, vendor.get_volume_id())
        return vendor.get_volume_name(), url

    def _get_parameter(self, func, run_safe=False, fail_return_value=''):
        if not run_safe:
            return func()
        try:
            return func()
        except:
            return fail_return_value

    def build_row(self, infinidat_device):
        from infi.vendata.powertools.utils import has_filesystem, is_mounted
        from infi.vendata.powertools.volume_list.engine import StorageModelDevice
        storage_model_device = StorageModelDevice(infinidat_device)
        volume_name, volume_url = self._get_volume_element(infinidat_device)
        # if a volume name is empty - it might be unmapped, so try to get only available info
        run_safe = False
        if volume_name == '':
            run_safe = True
        return dict(volume_name=volume_name,
                    volume_url=volume_url,
                    device_name=infinidat_device.get_display_name(),
                    serial=infinidat_device.get_scsi_serial_number(),
                    mount_point=self._get_parameter(storage_model_device.get_mount_point, run_safe, ''),
                    volume_size=self._get_parameter(infinidat_device.get_size_in_bytes, run_safe, 0),
                    paths=len(self._get_parameter(storage_model_device.get_connections, run_safe, [])),
                    lun=storage_model_device.get_lun(),
                    has_filesystem=False if is_veritas_multipathing_installed() else has_filesystem(infinidat_device),
                    is_mounted=is_mounted(infinidat_device),
                    is_multipath=storage_model_device.is_multipath(),
                    is_veritas=is_veritas_multipathing_installed())

    def get_mapped_devices(self, system_serial):
        from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices
        return set([device for device in get_infinidat_block_devices()
                    if self.is_device_belongs_to_this_system(device, system_serial)])

    def build_rows(self, system_serial):
        all_infinidat_devices = self.get_mapped_devices(system_serial)
        rows = [self.build_row(infinidat_device) for infinidat_device in all_infinidat_devices]
        return rows

    def get_row_data(self, system_serial):
        return self.build_rows(system_serial)


class BoxVolumes(object):
    """
    List all volumes in an infinibox that can be mapped
    """

    MAX_VOLUMES_TO_DISPLAY = 7  # this is what fits into the modal

    def get_total_volumes_in_system(self, system):
        """ returns the total number of volumes and clones in the system """
        params = dict(type='in:MASTER,CLONE', page_size=1)
        response = system.api.get('volumes?' + urlencode(params))
        return response.get_metadata()['number_of_objects']

    def search_system_volumes(self, serial, filter_string, show_clones, show_mapped, page):
        from infi.vendata.powertools.utils.queries import get_system_object_and_host
        system, host = get_system_object_and_host(serial)
        if system is None:
            return 0, 0, []
        params = dict(fields='id,name,type,pool_id',
                      page_size=self.MAX_VOLUMES_TO_DISPLAY, page=page, sort='name')
        if not show_mapped:
            params["mapped"] = "eq:false"
        else:
            mapped_volume_ids = [str(lun.volume_id) for lun in host.get_luns()]
            if len(mapped_volume_ids):
                params["id"] = "notin:" + ",".join(mapped_volume_ids)
        params['type'] = 'in:MASTER,CLONE' if show_clones else 'eq:MASTER'
        administered_pools = system.pools.get_administered_pools()
        if administered_pools:
            params['pool_id'] = 'in:{}'.format(','.join([str(pool.get_id()) for pool in administered_pools]))
        else:
            return self.get_total_volumes_in_system(system), 0, []
        if filter_string:
            params['name_contains'] = 'eq:' + filter_string
        response = system.api.get('volumes?' + urlencode(params))
        volumes = response.get_result()
        for volume in volumes:
            volume['pool_name'] = system.pools.get(id=volume['pool_id']).get_name()
        return self.get_total_volumes_in_system(system), response.get_metadata()['number_of_objects'], volumes

    @json_response
    def index(self, serial, show_mapped=False, **kwargs):
        """ returns aaData for dataTables, containing volumes in an infinibox. Supports
        filtering by text, hiding clone volumes, and uses the search API + paging mechanism to be optimal.
        returns maximum MAX_VOLUMES_TO_DISPLAY volumes, and the total count of volumes """
        draw = int(kwargs.get('draw'))
        display_start = int(kwargs.get('start'))
        dislpay_count = int(kwargs.get('length'))
        filter_string = kwargs.get('search[value]')
        show_clones = kwargs.get("show_clones", "false") != "false"
        radiobox_template = '<input type="radio" name="map-modal-volume-id-radio" value="{}"/>'
        icon_template = '<img src="/static/images/{}.png" class="map-vol-icon">'
        row_format = '<label class="map-modal-volume-row">{} {} / {}</label> {}'
        page = (display_start / dislpay_count) + 1
        total_volumes_in_system, total_matches, volumes = self.search_system_volumes(int(serial), filter_string, show_clones, show_mapped, page)
        table = []
        for vol in volumes:
            radiobox = radiobox_template.format(vol['id'])
            icon = icon_template.format("icon-clone" if vol['type'] == 'CLONE' else "icon-volume")
            row = row_format.format(radiobox, vol['pool_name'], vol['name'], icon)
            table.append([row])
        return dict(data=table,
                    recordsFiltered=total_matches,                 # total volumes that match
                    recordsTotal=total_volumes_in_system,          # total volumes overall
                    draw=draw)                                     # required by dataTables

class Volumes(object):
    """ list volumes mapped to this host """
    @json_response
    def index(self, serial, **kwargs):
        table_builder = HostVolumesTable()
        return table_builder.get_row_data(int(serial))

    performance = Performance()
    box = BoxVolumes()
