from ...case import Case, Category
from infi.hbaapi import get_ports_collection
from infi.pyutils.lazy import cached_method
from infi.storagemodel.vendor.infinidat.infinibox.wwn import InfinidatWWN, InvalidInfinidatWWN

from logging import getLogger
logger = getLogger(__name__)

class HelpingMixin(object):
    @cached_method
    def _is_port_of_infinidat(self, remote_port):
        try:
            InfinidatWWN(remote_port.port_wwn)
            return True
        except InvalidInfinidatWWN:
            return False

    @cached_method
    def _is_soft_target_of_infinidat(self, remote_port):
        try:
            return InfinidatWWN(remote_port.port_wwn).is_soft_target()
        except InvalidInfinidatWWN:
            return False

    def _is_port_connected_to_infinidat_soft_target(self, port):
        return any([self._is_soft_target_of_infinidat(remote_port) for remote_port in port.discovered_ports])

    def _is_port_connected_to_infinidat_storage(self, port):
        return any([self._is_port_of_infinidat(remote_port) for remote_port in port.discovered_ports])

    def _is_port_connected_to_other_systems(self, port):
        return any([not self._is_port_of_infinidat(remote_port) for remote_port in port.discovered_ports])

    def _get_infinibox_name_from_port(self, remote_port):
        from infi.storagemodel.base.scsi import SCSIDevice
        from infi.storagemodel.base.multipath import MultipathBlockDevice, MultipathStorageController
        from infi.storagemodel.connectivity import FCConnectivity

        def _get_system_name(device):
            paths = [device, ] \
                    if isinstance(device, SCSIDevice) else \
                    device.get_paths() \
                    if isinstance(device, (MultipathBlockDevice, MultipathStorageController)) else \
                    []
            for path in paths:
                if isinstance(path.get_connectivity(), FCConnectivity) and path.get_connectivity().get_target_wwn() == remote_port.port_wwn:
                    return device.get_vendor().get_system_name()

        try:
            from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices_and_controllers__mapped_to_lun0
            for device in get_infinidat_block_devices_and_controllers__mapped_to_lun0():
                name = _get_system_name(device)
                if name:
                    return name
        except:
            logger.exception("failed to deduce system name from remote port {}".format(remote_port.port_wwn))

    def _get_infinibox_serial_from_port(self, remote_port):
        return InfinidatWWN(remote_port.port_wwn).get_system_serial()

    def _get_infinibox_server_id_from_port(self, remote_port):
        return InfinidatWWN(remote_port.port_wwn).get_node_id()

    def _get_local_remote_port_tuple_to_infinibox(self):
        for port in get_ports_collection().get_ports():
            for remote_port in port.discovered_ports:
                if self._is_port_of_infinidat(remote_port):
                    yield port, remote_port

    def _get_initiators_by_infinibox_serial_and_name(self):
        result = {}
        for port, remote_port in self._get_local_remote_port_tuple_to_infinibox():
            serial = self._get_infinibox_serial_from_port(remote_port)
            name = self._get_infinibox_name_from_port(remote_port)
            connected_initiators = result.setdefault((serial, name), [])
            if port not in connected_initiators:
                connected_initiators.append(port)
        return result

    def _get_server_ids_by_infinibox_serial_and_name(self):
        result = {}
        for _, remote_port in self._get_local_remote_port_tuple_to_infinibox():
            # Exclude soft target ports since it doesn't have server (node) id...
            if not self._is_soft_target_of_infinidat(remote_port):
                serial = self._get_infinibox_serial_from_port(remote_port)
                name = self._get_infinibox_name_from_port(remote_port)
                server_id = self._get_infinibox_server_id_from_port(remote_port)
                connected_servers = result.setdefault((serial, name), [])
                if server_id not in connected_servers:
                    connected_servers.append(server_id)
        return result

    def _get_lun0_devices_count_by_infinibox_serial_and_name(self):
        from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices_and_controllers__mapped_to_lun0
        result = {}
        for device in get_infinidat_block_devices_and_controllers__mapped_to_lun0():
            serial = device.get_vendor().get_system_serial()
            name = device.get_vendor().get_system_name()
            count = result.get((serial, name), 0)
            result[(serial, name)] = count + 1
        return result


class DedicatedInitiators(Case, HelpingMixin):
    def run_test(self):
        msg = "Initiator port {} is attached both to InfiniBox storage and other target ports"
        connected_to_infinibox = False
        for port in get_ports_collection().get_ports():
            if self._is_port_connected_to_infinidat_storage(port):
                connected_to_infinibox = True
                if self._is_port_connected_to_other_systems(port):
                    self.warn(msg.format(port.port_wwn))
        if not connected_to_infinibox:
            self.skip("no fiberchannel ports connected to InfiniBox")

    def get_description(self):
        return "Checking if initiator ports connected to InfiniBox are not connected to other target ports"

    def get_kb_number(self):
        return "202319352"


class ConnectivityLayoutInitiatorSide(Case, HelpingMixin):
    def run_test(self):
        msg = "The host is connected to InfiniBox {} through only one initiator: {}"
        connections = self._get_initiators_by_infinibox_serial_and_name()
        if not connections:
            self.skip("no fiberchannel ports connected to InfiniBox")
        for (serial, name), connected_initiators in connections.items():
            if len(connected_initiators) <= 1:
                self.warn(msg.format("{} ({})".format(name, serial) if name else serial, connected_initiators[0].port_wwn))

    def get_description(self):
        return "Checking that each InfiniBox is connected through more than one initiator"

    def get_kb_number(self):
        return "202404261"


class ConnectivityLayoutTargetSide(Case, HelpingMixin):
    def run_test(self):
        msg = "The host is connected to only one node of InfiniBox {}: {}"
        connections = self._get_server_ids_by_infinibox_serial_and_name()
        if not connections:
            self.skip("no fiberchannel ports connected to a physical port of InfiniBox")
        for (serial, name), connected_servers in connections.items():
            if len(connected_servers) <= 1:
                self.warn(msg.format("{} ({})".format(name, serial) if name else serial, connected_servers[0]))

    def get_description(self):
        return "Checking that the host is connected to at least one Node of each InfiniBox"

    def get_kb_number(self):
        return "202404271"


class TooManyPaths(Case, HelpingMixin):
    MAX_PATHS_SUPPORTED = 32
    MAX_PATHS_RECOMMENDED = 12

    def run_test(self):
        msg = "The host has too many paths ({}) to InfiniBox {}, {}"
        connections = self._get_lun0_devices_count_by_infinibox_serial_and_name()
        if not connections:
            self.skip("no fiberchannel ports connected to InfiniBox")
        for (serial, name), lun_devices_count in connections.items():
            if lun_devices_count > self.MAX_PATHS_SUPPORTED:
                self.fail(msg.format(lun_devices_count, "{} ({})".format(name, serial) if name else serial,
                                     "maximum supported is {}".format(self.MAX_PATHS_SUPPORTED)))
            elif lun_devices_count > self.MAX_PATHS_RECOMMENDED:
                self.warn(msg.format("{} ({})".format(name, serial) if name else serial,
                                     lun_devices_count, "the recommended number is {}".format(self.MAX_PATHS_RECOMMENDED)))

    def get_description(self):
        return "Checking that the host does not have more block devices than the recommended limit"

    def get_kb_number(self):
        return "202319172"


def _try_storagemodel(func):
    from infi.pyutils.decorators import wraps
    @wraps(func)
    def decorated(*args, **kwargs):
        from infi.storagemodel.errors import StorageModelError
        try:
            return func(*args, **kwargs)
        except StorageModelError:
            logger.debug("caught StorageModelError", exc_info=True)
    return decorated


class TooManyVolumesMapped(Case, HelpingMixin):
    MAX_VOLUMES_ALLOWED = 32

    @_try_storagemodel
    def _get_count_of_multipath_block_devices(self):
        return len(self._model.get_native_multipath().get_all_multipath_block_devices())

    @_try_storagemodel
    def _get_count_of_unique_scsi_block_devices(self):
        devices = self._model.get_scsi().get_all_scsi_block_devices()
        return len({device.get_scsi_serial_number() for device in devices})

    def _get_count_of_unique_block_devices_count(self):
        # the purpose of this test is to see if there are too many LUNs provisioned
        # having too many luns can create aload on the scsi subsystem
        # which increases boot time and extends the error recovery period
        # this is why we want to first check all devices, and not just infinidat
        first_option = self._get_count_of_multipath_block_devices()
        # first_option can be None in case of error, or 0 if it's simply not configured
        # we want to return None only if there's an error, but if it's 0, we want to
        # calculate the unique scsi block devices using the other function
        if first_option is None:
            return None
        return first_option or self._get_count_of_unique_scsi_block_devices()

    def run_test(self):
        from infi.storagemodel import get_storage_model
        self._model = get_storage_model()
        msg = "The host has too many ({}) multipathed block devices mapped to it, consider changing volume mappings (recommended no more than {})"
        count = self._get_count_of_unique_block_devices_count()
        if count is None:
            self.skip("could not determine how many multipathed devices in total are attached to this host")
        if count > self.MAX_VOLUMES_ALLOWED:
            self.warn(msg.format(count, self.MAX_VOLUMES_ALLOWED))

    def get_description(self):
        return "Checking that the host does not have more volumes than the recommended number"

    def get_kb_number(self):
        return "202319182"


def get_category():
    return Category("Connectivity", [DedicatedInitiators, ConnectivityLayoutInitiatorSide,
                                     ConnectivityLayoutTargetSide, TooManyPaths, TooManyVolumesMapped])
