from infi.unittest import TestCase
from infi import unittest
from munch import Munch
from mock import patch
from .. import connectivity
from ... import case

INFINIBOX_HARD_TARGET_WWN = "5742b0f0004e2b11"
INFINIBOX_SOFT_TARGET_WWN = "2004742b0f00755b"

class DedicatedInitiatorsTestCase(TestCase):
    def test_there_are_no_ports(self):
        with patch.object(connectivity, "get_ports_collection") as get_ports_collection:
            get_ports_collection().get_ports.return_value = []
            with self.assertRaises(case.Skip):
                connectivity.DedicatedInitiators(None).run_test()

    @unittest.parameters.iterate("port_wwn_tested", [INFINIBOX_HARD_TARGET_WWN, INFINIBOX_SOFT_TARGET_WWN])
    def test_no_dedicated_ports(self, port_wwn_tested):
        with patch.object(connectivity, "get_ports_collection") as get_ports_collection:
            port = Munch(port_wwn="01"*8, port_name="xxxx",
                         discovered_ports=[Munch(port_wwn=port_wwn_tested), Munch(port_wwn="02"*8)])
            get_ports_collection().get_ports.return_value = [port]
            with self.assertRaises(case.Warning):
                connectivity.DedicatedInitiators(None).run_test()

    @unittest.parameters.iterate("port_wwn_tested", [INFINIBOX_HARD_TARGET_WWN, INFINIBOX_SOFT_TARGET_WWN])
    def test_dedicated_ports(self, port_wwn_tested):
        with patch.object(connectivity, "get_ports_collection") as get_ports_collection:
            port = Munch(port_wwn="01"*8, port_name="xxxx",
                         discovered_ports=[Munch(port_wwn=port_wwn_tested)])
            get_ports_collection().get_ports.return_value = [port]
            connectivity.DedicatedInitiators(None).run_test()


class ConnectivityLayoutInitiatorSide(TestCase):
    def test_there_are_no_ports(self):
        with patch.object(connectivity, "get_ports_collection") as get_ports_collection:
            get_ports_collection().get_ports.return_value = []
            with self.assertRaises(case.Skip):
                connectivity.ConnectivityLayoutInitiatorSide(None).run_test()

    def test_no_targets(self):
        with patch.object(connectivity, "get_ports_collection") as get_ports_collection:
            get_ports_collection().get_ports.return_value = [Munch(port_wwn="01"*8, discovered_ports=[])]
            with self.assertRaises(case.Skip):
                connectivity.ConnectivityLayoutInitiatorSide(None).run_test()

    @unittest.parameters.iterate("port_wwn_tested", [INFINIBOX_HARD_TARGET_WWN, INFINIBOX_SOFT_TARGET_WWN])
    def test_one_initiator_port(self, port_wwn_tested):
        infinibox_port = Munch(port_wwn=port_wwn_tested)
        with patch.object(connectivity, "get_ports_collection") as get_ports_collection:
            get_ports_collection().get_ports.return_value = [Munch(port_wwn="01"*8, discovered_ports=[infinibox_port])]
            with self.assertRaises(case.Warning):
                connectivity.ConnectivityLayoutInitiatorSide(None).run_test()

    @unittest.parameters.iterate("port_wwn_tested", [INFINIBOX_HARD_TARGET_WWN, INFINIBOX_SOFT_TARGET_WWN])
    def test_two_initiator_ports(self, port_wwn_tested):
        infinibox_port = Munch(port_wwn=port_wwn_tested)
        with patch.object(connectivity, "get_ports_collection") as get_ports_collection:
            get_ports_collection().get_ports.return_value = [Munch(port_wwn="01"*8, discovered_ports=[infinibox_port]),
                                                             Munch(port_wwn="02"*8, discovered_ports=[infinibox_port])]
            connectivity.ConnectivityLayoutInitiatorSide(None).run_test()


class ConnectivityLayoutTargetSideTestCase(TestCase):
    def test_there_are_no_ports(self):
        with patch.object(connectivity, "get_ports_collection") as get_ports_collection:
            get_ports_collection().get_ports.return_value = []
            with self.assertRaises(case.Skip):
                connectivity.ConnectivityLayoutTargetSide(None).run_test()

    def test_no_targets(self):
        with patch.object(connectivity, "get_ports_collection") as get_ports_collection:
            get_ports_collection().get_ports.return_value = [Munch(port_wwn="01"*8, discovered_ports=[])]
            with self.assertRaises(case.Skip):
                connectivity.ConnectivityLayoutTargetSide(None).run_test()

    def test_single_box_one_server(self):
        infinibox_port = Munch(port_wwn=INFINIBOX_HARD_TARGET_WWN)
        with patch.object(connectivity, "get_ports_collection") as get_ports_collection:
            get_ports_collection().get_ports.return_value = [Munch(port_wwn="01"*8, discovered_ports=[infinibox_port])]
            with self.assertRaises(case.Warning):
                connectivity.ConnectivityLayoutTargetSide(None).run_test()

    def test_single_box_three_servers(self):
        infinibox_ports = [Munch(port_wwn="5742b0f0004e2b{}1".format(index)) for index in [1, 2, 3]]
        with patch.object(connectivity, "get_ports_collection") as get_ports_collection:
            get_ports_collection().get_ports.return_value = [Munch(port_wwn="01"*8, discovered_ports=infinibox_ports)]
            connectivity.ConnectivityLayoutTargetSide(None).run_test()

    def test_two_boxes_one_good_one_bad(self):
        infinibox_ports = [Munch(port_wwn="5742b0f0004e2b{}1".format(index)) for index in [1, 2, 3]] + \
                          [Munch(port_wwn="5742b0f0004e2c11")]
        with patch.object(connectivity, "get_ports_collection") as get_ports_collection:
            get_ports_collection().get_ports.return_value = [Munch(port_wwn="01"*8, discovered_ports=infinibox_ports)]
            with self.assertRaises(case.Warning):
                connectivity.ConnectivityLayoutTargetSide(None).run_test()

    def test_two_good_boxes(self):
        infinibox_ports = [Munch(port_wwn="5742b0f0004f5b{}1".format(index)) for index in [1, 2, 3]]
        infinibox_ports += [Munch(port_wwn="5742b0f0004f5c{}1".format(index)) for index in [1, 2, 3]]
        with patch.object(connectivity, "get_ports_collection") as get_ports_collection:
            get_ports_collection().get_ports.return_value = [Munch(port_wwn="01"*8, discovered_ports=infinibox_ports)]
            connectivity.ConnectivityLayoutTargetSide(None).run_test()


class TooManyPathsTestCase(TestCase):
    def test_few_paths(self):
        from capacity import GB, TiB
        from infi.vendata.smock import HostMock
        mock = HostMock()
        inventory = mock.get_inventory()
        inventory.add_initiator()
        system = inventory.add_infinibox(9999)
        system.pools.create(name="default", virtual_capacity=100 * TiB, physical_capacity=100 * TiB)
        inventory.zone_with_system__best_practices(system)
        inventory.register_host_in_system(system)
        with mock:
            inventory.provision_a_volume_from_system(system, 123 * GB)
            connectivity.TooManyPaths(None).run_test()

    def test_too_many_paths(self):
        from capacity import GB, TiB
        from infi.vendata.smock import HostMock
        mock = HostMock()
        inventory = mock.get_inventory()
        inventory.add_initiator()
        inventory.add_initiator()
        system = inventory.add_infinibox(9999)
        system.pools.create(name="default", virtual_capacity=100 * TiB, physical_capacity=100 * TiB)
        inventory.zone_with_system__full_mesh(system)
        inventory.register_host_in_system(system)
        with mock:
            inventory.provision_a_volume_from_system(system, 123 * GB)
            with self.assertRaises(case.Failure):
                connectivity.TooManyPaths(None).run_test()
