from infi.vendata.powertools.settings.case import FixableCase, Case, Skip, Error, Category
from infi.pyutils.decorators import wraps
from infi.storagemodel import errors
from infi.storagemodel.vendor.infinidat import shortcuts
from infi.vendata.powertools.utils import get_infinidat_non_multipath_block_devices
from logging import getLogger
logger = getLogger(__name__)


def raise_skip_on_any_storage_related_error(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except errors.StorageModelError:
            raise Skip("storagemodel error")
    return wrapper


@errors.check_for_scsi_errors
def get_reported_luns(device):
    from infi.asi.cdb.report_luns import ReportLunsCommand
    from infi.asi.coroutines.sync_adapter import sync_wait
    command = ReportLunsCommand(select_report=0)
    with device.asi_context() as asi:
        return sync_wait(command.execute(asi)).lun_list


def get_hctl_sets_for_device(device):
    from infi.storagemodel.base.multipath import MultipathDevice
    if isinstance(device, MultipathDevice):
        return set([path.get_hctl() for path in device.get_paths()])
    return set([device.get_hctl()])


class MultipathDeviceForVolume(Case):
    @classmethod
    def get_non_multipathed_devices_map(cls):
        device_map = dict()
        for d in get_infinidat_non_multipath_block_devices():
            if d.get_scsi_standard_inquiry().peripheral_device.qualifier == 0:
                # in linux, the device might get disconnected before rescan, and in this case
                # we might still have the scsi device but not the multipath device
                volume_name = d.get_vendor().get_volume_name()
                # to deal with the case the device disappears while we are in this iteration, we do the following
                # empty name means that we still see the SCSI device fro the LuN, but there is no volume mapped
                try:
                    if not d.get_scsi_serial_number():
                        continue
                except:
                    continue
                if volume_name:
                    device_map[d.get_vendor().get_volume_name()] = d.get_display_name()
        return device_map

    @raise_skip_on_any_storage_related_error
    def run_test(self):  # pragma: no cover
        message_template = "Found volumes that don't have a multipath device (Volumes: {})"
        devices = self.get_non_multipathed_devices_map().keys()
        self.assertEqual(devices, list(), message_template.format(", ".join([str(item) for item in devices])))

    def get_description(self):
        return "Checking for volumes that don't have a multipath device"

    def get_kb_number(self):
        return "202319252"


class BlockDevicesMatchReportedLuns(FixableCase):
    @classmethod
    def get_actual_hctl_set_for_infinibox_block_devices(cls):
        all_infinibox_block_devices = shortcuts.get_infinidat_block_devices()
        actual_hctl_set = set()
        for device in all_infinibox_block_devices:
            if device.get_scsi_standard_inquiry().peripheral_device.qualifier == 0:
                # in linux, the device might get disconnected before rescan, and in this case
                # we might still have the scsi device but not the multipath device
                actual_hctl_set.update(get_hctl_sets_for_device(device))
        return actual_hctl_set

    @classmethod
    def get_expected_hctl_set_for_infinibox_block_devices(cls):
        from infi.storagemodel.base.scsi import SCSIBlockDevice
        from infi.storagemodel.base.multipath import MultipathBlockDevice
        all_infinibox_lun0_devices = shortcuts.get_infinidat_block_devices_and_controllers__mapped_to_lun0()
        expected_set = set()
        for device in all_infinibox_lun0_devices:
            for hctl in get_hctl_sets_for_device(device):
                expected_set.update(set([hctl[lun] for lun in get_reported_luns(device) if
                                         lun != 0 or isinstance(device, (SCSIBlockDevice, MultipathBlockDevice))]))
        return expected_set

    @raise_skip_on_any_storage_related_error
    def run_test(self):
        actual = self.get_actual_hctl_set_for_infinibox_block_devices()
        expected = self.get_expected_hctl_set_for_infinibox_block_devices()
        error_suffix = "; please collect logs and contact support"
        if not actual.issubset(expected):
            error_template = "actual HCTLs for block devices: {} - are not a subset of the expected HCTLs from ReportLuns"
            error_message = error_template.format(", ".join([str(d) for d in list(actual - expected)]))
            logger.exception(error_message)
            raise Error("{}{}".format(error_message, error_suffix))
        message_template = "InfiniBox block devices are missing for the following HCTLS (after rescan): {}{}"
        self.assertEquals(actual, expected, message_template.format(expected - actual, error_suffix))

    def run_fix(self):
        from infi.vendata.powertools.rescan.engine import do_rescan
        do_rescan(disable_print=True)

    def get_description(self):
        return "Checking that all reported INFINIDAT LUNs have a block device"

    def get_kb_number(self):
        return "202404201"

class PathCount(Case):
    @classmethod
    def get_number_of_fc_connections_to_each_infinibox(cls):
        serial_number_count = dict()
        for system_serial in cls.iter_infinibox_system_serial_numbers_from_hbaapi():
            serial_number_count[system_serial] = serial_number_count.setdefault(system_serial, 0) + 1
        return serial_number_count

    @classmethod
    def iter_infinibox_system_serial_numbers_from_hbaapi(cls):
        from infi.hbaapi import get_ports_collection
        from infi.storagemodel.vendor.infinidat.infinibox.wwn import InfinidatWWN, InvalidInfinidatWWN
        for initiator_port in get_ports_collection().iter_ports():
            for target_port in initiator_port.discovered_ports:
                try:
                    yield InfinidatWWN(target_port.port_wwn).get_system_serial()
                except InvalidInfinidatWWN:
                    pass

    @raise_skip_on_any_storage_related_error
    def run_test(self):
        expected_by_system_serial = self.get_number_of_fc_connections_to_each_infinibox()
        for device in shortcuts.get_infinidat_native_multipath_block_devices():
            actual = len(device.get_paths())
            system_serial = device.get_vendor().get_system_serial()
            expected = expected_by_system_serial.get(system_serial)
            if not expected:
                error_template = "no fc connections found to multipath device {} from system {}"
                error_message = error_template.format(device, system_serial)
                logger.exception(error_message)
                raise Error("{}; please collect logs and contact support".format(error_message))
            if actual > expected:
                error_template = "Path count for device {} is higher than fc connections to system {}"
                error_message = error_template.format(device, system_serial)
                raise Error("{}; please collect logs and contact support".format(error_message))
            message_template = "InfiniBox device {} has less paths than expected ({}<{})"
            self.assertEqual(actual, expected, message_template.format(device, actual, expected))

    def get_description(self):
        return "Checking the path count for all INFINIDAT MPIO devices"

    def get_kb_number(self):
        return "202404341"


def get_category():
    return Category("Devices", [MultipathDeviceForVolume, BlockDevicesMatchReportedLuns, PathCount])
