from ...case import Case, Category, Warning
from infi.pyutils.lazy import cached_method
from infi.hbaapi import get_ports_collection
from infi.os_info.parse_version import parse_version
from logging import getLogger

logger = getLogger(__name__)

# pylint: disable=R0904

KNOWN_QLOGIC_MODELS = ['QLA2342', 'QLE2462', 'QLE2464', 'QLE2562', 'QLE2564', 'HPAJ764A']

# http://www.qlogic.com/Resources/Documents/FAQs/Adapters/brocade-1860-fabric-adapter-external-faqs.pdf
# http://driverdownloads.qlogic.com/QLogicDriverDownloads_UI/DefaultNewSearch.aspx
KNOWN_BROCADE_QLOGIC_MODELS = ['QLogic-815', 'QLogic-825',
                               'QLogic-1860-1p', 'QLogic-1860-2p',
                               'QLogic-1860-1f', 'QLogic-1860-2f',
                               'QLogic-1860-1c', 'QLogic-1860-2c',
                               ]

def get_fc_hba_company_from_port(port):
    """
    :returns: "Emulex", "QLogic", "Cisco", "Brocade" or None
    """

    # special cases
    if port.model and 'QLogic' in str(port.model) and port.model_description and 'BR-series' in str(port.model_description):
        # Broacade HBAs with new firmware are identified as QLogic, but use the Brocade driver
        return "Brocade"
    if port.port_symbolic_name and 'fnic' in str(port.port_symbolic_name):
        return "Cisco"

    # check model, model_description and manufacturer
    for port_str in port.model, port.model_description, port.manufacturer, port.port_symbolic_name:
        if port_str:  # either None or an empty string
            port_str = port_str.split()[0]
            if port_str in KNOWN_QLOGIC_MODELS:
                return "QLogic"
            if port_str in KNOWN_BROCADE_QLOGIC_MODELS:
                return "Brocade"
            if port_str in ["QLogic", "Emulex", "Brocade", "Cisco"]:
                return port_str

    return None


class SupportedFiberChannelHostBusAdapters(Case):
    def run_test(self):
        ports = get_ports_collection().get_ports()
        logger.info("ports retrieved from hbaapi: {}".format(ports))
        if not ports:
            self.skip("no fiberchannel ports detected")
        for port in ports:
            if get_fc_hba_company_from_port(port) is None:
                self.warn("There are FC HBAs installed from an unsupported manufacturer: [{}] {}".format(port.model, port.model_description))

    def get_description(self):
        return "Looking for supported Fibre Channel host bus adapters"

    def get_kb_number(self):
        return "202404231"


class SupportedFiberChannelDrivers(Case):
    @cached_method
    def _get_dist_name(self):
        from infi.vendata.powertools.utils import get_distribution, get_platform_name
        dist = get_platform_name()
        return dist if dist != 'linux' else get_distribution().distname

    def _get_major_version(self):
        return getattr(self, "_get_{}_version".format(self._get_dist_name()))()

    def _get_redhat_version(self):
        from infi.vendata.powertools.utils import get_distribution
        return get_distribution().version.split('.')[0]

    def _get_centos_version(self):
        return self._get_redhat_version()

    def _get_ubuntu_version(self):
        from infi.vendata.powertools.utils import get_distribution
        return get_distribution().version

    def _get_suselinux_version(self):
        from infi.vendata.powertools.utils import get_distribution
        return get_distribution().version

    def _get_windows_version(self):
        from infi.winver import Windows
        return Windows().version

    def _get_sunos_version(self):
        from infi.os_info import get_platform_string
        return get_platform_string().split('-')[1]

    def _get_case_config(self):
        from infi.vendata.powertools.conf import get_configuration
        return get_configuration().get('readiness.cases.fc_driver_version')

    def _validate_configuration_exists(self, company):
        if company not in self._get_case_config() or \
           self._get_major_version() not in self._get_case_config()[company]["MINIMAL"][self._get_dist_name()]:
            raise Warning("fc vendor {} has not been tested for this os version".format(company))

    def _choose_minimal_version(self, company):
        return self._get_case_config()[company]["MINIMAL"][self._get_dist_name()][self._get_major_version()]

    def _choose_recommended_version(self, company):
        return self._get_case_config()[company]["RECOMMENDED"][self._get_dist_name()][self._get_major_version()]

    def _get_incompatible_versions(self, company):
        return self._get_case_config()[company]["INCOMPATIBLE"][self._get_dist_name()]

    def _apply_version_workarounds(self, version):
        # So drivers have weird behvaiours, this is where we work around them
        # Emulex Windows OEM drivers append 7-, 7. or 5. to the start of their version number
        # This is probably a bad workaround, but I don't see another way currently.
        if version.startswith('7-') or version.startswith('7.') or version.startswith('5.'):
            return version[2:]
        if version.endswith('-debug'):
            return version.replace('-debug', '')
        return version.replace('v.', '').strip('()')

    def run_test(self):
        ports = get_ports_collection().get_ports()
        if not ports:
            self.skip("no fiberchannel ports detected")
        for port in ports:
            company = get_fc_hba_company_from_port(port)
            if company is None:
                continue
            existing_version = self._apply_version_workarounds(port.driver_version.split()[-1])
            self._validate_configuration_exists(company)
            minimal_version = self._choose_minimal_version(company)
            recommended_version = self._choose_recommended_version(company)
            incompatible_versions = self._get_incompatible_versions(company)
            parsed_existing = parse_version(existing_version)
            if minimal_version == "0":
                self.skip("Allowing any fiberchannel driver version on this platform")
            if not parsed_existing >= parse_version(minimal_version):
                msg = "{} driver version is lower than required: {} < {}"
                self.warn(msg.format(company, existing_version, minimal_version))
            if any(parsed_existing == parse_version(item) for item in incompatible_versions):
                msg = "{} driver version {} is NOT compatible. Recommended version is {}"
                self.fail(msg.format(company, existing_version, recommended_version))
            if not parsed_existing >= parse_version(recommended_version):
                msg = "{} driver version is lower than recommended: {} < {}"
                self.warn(msg.format(company, existing_version, recommended_version))

    def get_description(self):
        return "Checking versions of FC HBA drivers"

    def get_kb_number(self):
        return "202319342"

def get_category():
    return Category("FC HBA", [SupportedFiberChannelHostBusAdapters, SupportedFiberChannelDrivers])
