from infi.vendata.powertools.settings.case import FixableCase
from infi.pkgmgr import RedHatPackageManager
from infi.vendata.powertools.settings.cases.linux.base_cases import SoftwareDependency
from infi.vendata.powertools.utils import execute_command
from infi.reboot import ask_for_reboot, has_reboot_took_place
from logging import getLogger

REBOOT_FLAG = 'regenerate-initrd'
logger = getLogger(__name__)


def iter_lvm_underlying_block_devices(block_access_path):
    try:
        #  /dev/vg_sap02/lv_root:vg_sap02:3:1:-1:1:16547840:2020:-1:0:-1:253:0
        lvdisplay = execute_command(["lvdisplay", "-c", block_access_path])
        #  /dev/sdi2:vg_sap02:lvm2:a--:8.82g:0
        pvs = execute_command(["pvs", "--noheadings", "--nosuffix", "--separator=:"])
        vg_name = lvdisplay.get_stdout().strip().splitlines()[0].split(':')[1]
        for line in pvs.get_stdout().splitlines():
            _pv, _vg, _ = line.strip().split(':', 2)
            if _vg == vg_name:
                yield _pv
    except:
        logger.exception("iter_lvm_underlying_block_devices({!r}) failed".format(block_access_path))


def does_device_or_its_partitions_matches_block_access_path(device, block_access_path):
    from infi.storagemodel.base.disk import NoSuchDisk, NoPartitionTable
    if device.get_block_access_path() == block_access_path:
        logger.debug("{!r} is the block device for the root filesystem".format(device))
        return True
    try:
        for partition in device.get_disk_drive().get_partition_table().get_partitions():
            if partition.get_block_access_path() == block_access_path:
                logger.debug("{!r} is the partition of block device {!r} for the root filesystem".format(partition, device))
                return True
    except (NoSuchDisk, NoPartitionTable):
        pass
    except:
        # HPT-1666
        # On some systems, parted returned a weird error (No medium found) and it is correct to raise an exception about it from infi.parted
        # So we catch it here and move on
        pass
    return False


def get_storage_model_devices_for_root_filesystem():
    from infi.storagemodel import get_storage_model
    mounts = get_storage_model().get_mount_manager().get_mounts()
    root_filesystems = [mount for mount in mounts if
                        mount.get_mount_point() == '/']
    if len(root_filesystems) != 1:
        # fix for HPT-1636
        root_filesystems = [mount for mount in root_filesystems if
                            mount.get_block_access_path() != 'rootfs']
    if len(root_filesystems) != 1:
        logger.error("failed to deduce mount for root filesystem (found these: {!r})".format(root_filesystems))
        return []

    get_scsi_devices = get_storage_model().get_scsi().get_all_scsi_block_devices
    get_multipath_devices = get_storage_model().get_native_multipath().get_all_multipath_block_devices
    for device in get_scsi_devices() + get_multipath_devices():
        if does_device_or_its_partitions_matches_block_access_path(device, root_filesystems[0].get_block_access_path()):
            return [device]

    # maybe its lvm, lets try to find out without storagemodel that does not support it
    block_access_path = root_filesystems[0].get_block_access_path()
    boot_devices = []
    for underlying_block_device in iter_lvm_underlying_block_devices(block_access_path):
        for device in get_scsi_devices() + get_multipath_devices():
            if does_device_or_its_partitions_matches_block_access_path(device, underlying_block_device):
                logger.debug("{!r} is one of the block devices for the root filesystem".format(device))
                boot_devices.append(device)

    if not boot_devices:
        logger.error("failed to locate storagemodel device for root filesystem {!r}".format(root_filesystems[0]))
    return boot_devices


def is_booting_from_san():
    from infi.storagemodel.base.multipath import MultipathDevice
    from infi.storagemodel.base.scsi import SCSIDevice
    from infi.storagemodel.connectivity import FCConnectivity
    try:
        for device in get_storage_model_devices_for_root_filesystem():
            connectivity = device.get_connectivity() if isinstance(device, SCSIDevice) else \
                           device.get_paths()[0].get_connectivity() if isinstance(device, MultipathDevice) else \
                           None
            if isinstance(connectivity, FCConnectivity):
                return True
        return False
    except:
        logger.exception("failed to deduce boot from san; if it is booting from SAN, its not from InfiniBox probably")
        return None


def is_regeneration_of_initrd_needed():
    return not has_reboot_took_place(REBOOT_FLAG)


def mark_generation_of_initrd_as_needed():
    ask_for_reboot(REBOOT_FLAG)


def is_ubuntu():
    from infi.os_info import get_platform_string
    return 'ubuntu' in get_platform_string()

def is_suse_12():
    from infi.os_info import get_platform_string
    return 'suse-12' in get_platform_string()

class GenerateInitScript(FixableCase):
    def run_test(self):
        if not is_regeneration_of_initrd_needed():
            self.skip("no boot-related changes were performed")
        # we made some changes that require re-generating initrd
        # usually, we need to check if the server boots from SAN to decide if
        # to generate new initrd image
        # one exception is ubuntu, which requires generating a new initrd anyway,
        # because it loads the fc driver at boot time without a trick to load parameters from modprobe.conf
        # that is on the root filesystem
        if is_ubuntu() or is_suse_12():
            self.fail("boot-related changes were performed; initrd must be regenerated")
        result = is_booting_from_san()
        if result is False:
            self.skip("server does not boot from SAN")
        if result is None:
            self.skip("failed to determine if the server boots from SAN; if it does, manually regenerate initrd for changes to take effect during boot")
        self.fail("boot-related changes were performed; initrd must be regenerated")

    def get_fix_description(self):
        return "save a backup initrd image generate a new one"

    def run_fix(self):
        from infi.vendata.powertools.settings.utils import backup_and_create_initrd
        backup_and_create_initrd()
        self.ask_for_reboot("Reboot is required for the changes to take effect")

    def get_description(self):
        return "checks if PowerTools has made some changes that require a new initrd image in boot-from-san environments"

    def get_kb_number(self):
        return "202621702"


class DracutPackage(SoftwareDependency):
    def __init__(self, category):
        super(DracutPackage, self).__init__(category, "202404001", "dracut", RedHatPackageManager)
