from infi.vendata.powertools.settings.case import FixableCase
from infi.vendata.powertools.utils import execute_command
from pkg_resources import resource_filename
import os

from logging import getLogger
logger = getLogger(__name__)


# pylint: disable=R0904,W0201

FC_HBA_DRIVER_ATTRIBUTES_DEFAULTS = {'lpfc': {'lpfc_hba_queue_depth': "4096",
                                              'lpfc_lun_queue_depth': "30",
                                              'lpfc_devloss_tmo': "30",
                                              'lpfc_log_verbose': '0'},
                                     'qla2xxx': {'ql2xmaxqdepth': "32",
                                                 'ql2xextended_error_logging': '0',
                                                 'qlini_mode': 'enabled'},
                                     'bfa': {
                                         'bfa_lun_queue_depth': '32',
                                         'rport_del_timeout': '90',
                                         # removed this one because the name varies on different driver versions
                                         # 'bfa_log_level': '3',
                                         'msix_disable_cb': '0',
                                         'msix_disable_ct': '0'}
                                     }

FC_HBA_DRIVER_NAME_TO_COMPANY = {'lpfc': 'Emulex',
                                 'qla2xxx': 'QLogic',
                                 'bfa': 'Brocade'}

def is_there_a_fc_driver_loaded():
    from infi.vendata.powertools.utils import execute_command
    lsmod = execute_command(["/sbin/lsmod"])
    output = lsmod.get_stdout() + lsmod.get_stderr()
    return any([module_name in output for module_name in FC_HBA_DRIVER_ATTRIBUTES_DEFAULTS.keys()])

def reloaded_fc_drivers():
    def _reload_module(module_name):
        from infi.vendata.powertools.settings.cases.linux.base_cases.kernel import KernelModule
        kernel_module = KernelModule(module_name)
        if kernel_module.is_loaded():
            try:
                kernel_module.unload()
                kernel_module.load()
            except RuntimeError:
                logger.exception("failed to reload {!r}".format(kernel_module))
                return False
        return True

    return all([_reload_module(module_name) for module_name in FC_HBA_DRIVER_ATTRIBUTES_DEFAULTS.keys()])

class FiberChannelPropertiesMixin(object):
    def _get_driver_attribute_from_sysfs(self, driver_name, key):
        filename = "/sys/module/{}/parameters/{}".format(driver_name, key)
        if not os.path.exists(filename):
            return None
        with open(filename) as fd:
            return fd.read().splitlines()[0].strip()

    def _get_drive_attribute_default_value(self, driver_name, key):
        return FC_HBA_DRIVER_ATTRIBUTES_DEFAULTS[driver_name][key]

    def _get_active_port_attribute(self, driver_name, key):
        from infi.hbaapi import get_ports_collection
        from infi.vendata.powertools.settings.cases.fc_driver_version import get_fc_hba_company_from_port
        parameters = []
        ports = get_ports_collection().get_ports()
        for port in ports:
            if FC_HBA_DRIVER_NAME_TO_COMPANY[driver_name] == get_fc_hba_company_from_port(port):
                port_host = port.hct[0]
                parameter_path = "/sys/class/scsi_host/host{}/{}".format(port_host, key)
                if os.path.exists(parameter_path):
                    with open(parameter_path) as fd:
                        parameter = fd.read().splitlines()[0].strip()
                        parameters.append(parameter)
        if len(parameters) == 0:
            return None
        elif len(set(parameters)) > 1:
            msg = "Found different values for parameter {} in active ports attributes: {}"
            self.warn(msg.format(key, parameters))
            return None
        return parameters[0]

    def _get_driver_attribute(self, driver_name, key):
        res = self._get_driver_attribute_from_sysfs(driver_name, key)
        if res is not None:
            return res
        # this is a hack for some versions of Emulax that doesn't keep the attributes in the module folder
        if driver_name == 'lpfc':
            return self._get_active_port_attribute(driver_name, key)
        return self._get_drive_attribute_default_value(driver_name, key)

    def _get_case_config(self):
        from infi.vendata.powertools.conf import get_configuration
        return get_configuration().get('readiness.cases.fc_driver_attributes')

    def _try_to_convert_to_int(self, n):
        try:
            return int(n) if n.isdigit() else int(n, 16) if n.startswith('0x') else n
        except:
            return n

    def _is_value_matches(self, actual, expected):
        from re import match
        trivial_result = bool(match(actual, expected))
        actual = self._try_to_convert_to_int(actual)
        expected = self._try_to_convert_to_int(expected)
        if isinstance(actual, (int, long)) and isinstance(expected, (int, long)):
            return actual == expected
        else:
            return trivial_result

    def _get_supported_params(self, driver_name):
        # we read the supported parameters straight from the ko binary - the parameters are listed in the
        # ".modinfo" section in that binary, it's easy to extract with a simple regex
        import os
        import re
        driver_path = os.path.join("/lib", "modules", os.uname()[2], "kernel", "drivers", "scsi", driver_name)
        driver_full_path = os.path.join(driver_path, driver_name + ".ko")
        driver_elf_data = open(driver_full_path, "rb").read()
        return re.findall("parm=(\w+)", driver_elf_data)


class FiberChannelDriverProperties(FixableCase, FiberChannelPropertiesMixin):
    def __init__(self, *args, **kwargs):
        super(FiberChannelDriverProperties, self).__init__(*args, **kwargs)
        from platform import linux_distribution
        linux_dist, linux_ver, _ = linux_distribution()
        ldist = linux_dist.lower()
        self.is_redhat = "red hat" in ldist or "centos" in ldist
        if linux_ver:
            self.linux_version = int(linux_ver.split(".")[0])
        self.modprobe_conf = os.path.join(os.path.sep, 'etc', 'modprobe.conf')
        self.modprobe_d = os.path.join(os.path.sep, 'etc', 'modprobe.d')

    def run_test(self):
        config = self._get_case_config()
        for driver_name in config.keys():
            if not os.path.exists("/sys/module/{}".format(driver_name)):
                continue
            supported_driver_parameters = self._get_supported_params(driver_name)
            for key, expected in config[driver_name].items():
                if key not in supported_driver_parameters:
                    continue
                msg = "Property {} of {} driver is {}, when it should be {}"
                value_in_conf = self._get_driver_attribute(driver_name, key)
                equals = self._is_value_matches(value_in_conf, expected)
                self.assertTrue(equals, msg.format(key, driver_name, value_in_conf, expected))
                msg = "Property {} of {} driver active port is {}, while it is configured to the recommended value of {}"
                value_in_port = self._get_active_port_attribute(driver_name, key)
                # if value in port can't be found - ignore
                if value_in_port is not None and not self._is_value_matches(value_in_port, expected):
                    self.warn(msg.format(key, driver_name, value_in_port, expected))

        if not self._should_write_to_modprobe_conf():
            if not os.path.exists(self.modprobe_d):
                self.warn("{} does not exist".format(self.modprobe_d))
            if not os.path.isdir(self.modprobe_d):
                self.fail("{} is not a directory".format(self.modprobe_d))

    def _should_write_to_modprobe_conf(self):
        return ((self.is_redhat and self.linux_version == 5) and os.path.exists(self.modprobe_conf)) or \
               ((self.is_redhat and self.linux_version in (5, 6)) and not os.path.isdir(self.modprobe_d))

    def _write_driver_parameters(self):
        # modprobe.d doesn't work in Redhat in all cases, and modprobe.conf doesn't work in Ubuntu...
        # linux at its best!
        import re
        import glob

        def get_current_configuration_data(output_file):
            if os.path.exists(output_file):
                return open(output_file, 'r').read()
            return ""

        def get_all_configuration_data():
            glob_file_ext = '*' if (self.is_redhat and self.linux_version == 5) else '*.conf'
            all_config_data = ''
            if not (self.is_redhat and self.linux_version == 5 and os.path.exists(self.modprobe_conf)):
                all_config_data += '\n'.join([open(f, 'r').read() for f in glob.glob(os.path.join(os.path.sep, 'etc', 'modprobe.d', glob_file_ext))])
            if (self.is_redhat and self.linux_version in (5, 6)) and os.path.exists(self.modprobe_conf):
                all_config_data += '\n' + open(self.modprobe_conf, 'r').read()

            if (self.is_redhat and self.linux_version == 5):
                include_pattern = re.compile("^include\s+([/\S]*)$", re.MULTILINE)
                include_filenames = include_pattern.findall(all_config_data)
                included_data = '\n'.join([open(f, 'r').read() for f in include_filenames if os.path.exists(f)])
                all_config_data += '\n' + included_data
                if include_pattern.search(included_data):
                    # I didn't want to handle weird cases such circular imports and etc..
                    self.warn("Nested includes in {} are not supported".format(self.modprobe_conf))
            return all_config_data

        def get_output_filename():
            # On redhat 5 we must make sure modprobe.conf exists, because if we create one and it wasn't there,
            # All the settings inside modprobe.d will be ignored!
            if self._should_write_to_modprobe_conf():
                output_file = self.modprobe_conf
            else:
                if not os.path.exists(self.modprobe_d):
                    os.mkdir(self.modprobe_d)
                if not os.path.isdir(self.modprobe_d):
                    self.fail("{} is not a directory".format(self.modprobe_d))
                output_file = os.path.join(self.modprobe_d, 'infinidat.conf')
                # remove our configuration file if it exists - in order to handle configuration changes made by us
                if os.path.exists(output_file):
                    os.remove(output_file)
            return output_file

        output_file = get_output_filename()
        current_config_data = get_current_configuration_data(output_file)
        all_config_data = get_all_configuration_data()
        config = self._get_case_config()
        with open(output_file, 'a') as fd:
            for driver_name in config.keys():
                if not os.path.exists("/sys/module/{}".format(driver_name)):
                    continue
                supported_driver_parameters = self._get_supported_params(driver_name)
                new_options = [(key, value) for key, value in config[driver_name].items()
                               if key in supported_driver_parameters]
                if len(new_options) == 0:
                    continue
                # we write one line with all options per driver - writing a line for each option doesn't always work
                new_options_chained = ' '.join(["{}={}".format(key, value) for key, value in new_options])
                new_options_line = "options {} {}\n".format(driver_name, new_options_chained)
                # make sure that the configuration line doesn't already exist
                if re.search("^{}$".format(new_options_line[:-1]), current_config_data, re.MULTILINE) is not None:
                    self.skip("Configuration already exist in configuration file")
                # make sure we don't have a conflict in the configuration
                for key, value in new_options:
                    regfind = re.search("^options {} (?:.* )?{}=(\S*)(?: .*)?$".format(driver_name, key), all_config_data, re.MULTILINE)
                    if regfind is not None:
                        if not str(regfind.groups()[0]) == value:
                            self.warn("There are conflicts in driver configuration. The string value of parameter {} is set to {!r} while we expected it to be {!r}".format(key, regfind.groups()[0], value))
                fd.write(new_options_line)

    def get_fix_description(self):
        return "Overrides the fiberchannel driver attributes"

    def run_fix(self):
        from infi.vendata.powertools.settings.cases.linux.base_cases.boot_from_san import mark_generation_of_initrd_as_needed
        self._write_driver_parameters()
        mark_generation_of_initrd_as_needed()
        if not reloaded_fc_drivers():
            self.ask_for_reboot("Reboot is required for the changes to take effect")

    def get_description(self):
        return "Checking FC HBA driver attributes"

    def get_kb_number(self):
        return "202319332"


class RcModulesScript(FixableCase):
    SRC = resource_filename(__name__, "rc.modules")
    DST = os.path.join(os.path.sep, 'etc', 'rc.modules.infinidat')
    DST_LINK = os.path.join(os.path.sep, 'etc', 'rc.modules')
    PERMISSIONS = 0755

    def _prepare_dst(self):
        with open(self.SRC) as fd:
            contents = fd.read()
        with open(self.DST, 'w') as fd:
            fd.write(contents)
        os.chmod(self.DST, self.PERMISSIONS)

    def _assert_link(self):
        self.assertTrue(os.path.exists(self.DST_LINK), '{!r} does not exist'.format(self.DST_LINK))
        self.assertTrue(os.path.islink(self.DST_LINK), '{!r} is not a link to {!r}'.format(self.DST_LINK, self.DST))
        self.assertTrue(os.readlink(self.DST_LINK) == self.DST, '{!r} is not a link to {!r}'.format(self.DST_LINK, self.DST))

    def run_test(self):
        if not is_there_a_fc_driver_loaded():
            self.skip("no need to install /etc/rc.modules since there is no fc driver that requires modification")
        self._prepare_dst()
        self._assert_link()

    def get_fix_description(self):
        return "Writes /etc/rc.modules.infinidat and sets /etc/rc.modules point to it"

    def run_fix(self):
        if os.path.exists(self.DST_LINK):
            if os.path.islink(self.DST_LINK) and os.readlink(self.DST_LINK) == self.DST:
                pass
            else:
                self.fail("{!r} alrady exists and does not point to {!r}. Review the file contents and create the link manually".format(self.DST_LINK, self.DST))
        else:
            os.symlink(self.DST, self.DST_LINK)

    def get_description(self):
        return "Checking that /etc/rc.modules is executable and points to /etc/rc.modules.infinidat"

    def get_kb_number(self):
        return "202696031"


class SystemdUnitFile(FixableCase):
    SERVICE_NAME = "infi.fc.modules.restart.service"
    SRC = resource_filename(__name__, SERVICE_NAME)
    DST_DIR = '/etc/systemd/system'
    DST_FILE = os.path.join(DST_DIR, SERVICE_NAME)

    def _prepare_dst(self):
        with open(self.SRC) as fd:
            contents = fd.read()
        with open(self.DST_FILE, 'w') as fd:
            fd.write(contents)

    def _assert_enabled(self):
        try:
            cmd = execute_command(["systemctl", "is-enabled", self.SERVICE_NAME], check_returncode=True)
        except RuntimeError:
            self.fail("service {} is not enabled".format(self.SERVICE_NAME))
        return "enabled" in cmd.get_stdout()

    def run_test(self):
        if not is_there_a_fc_driver_loaded():
            self.skip("no need to install unit file since there is no fc driver that requires modification")
        self._prepare_dst()
        self._assert_enabled()

    def get_fix_description(self):
        return "Writing {} unit file to {} and enables it".format(self.SERVICE_NAME, self.DST_DIR)

    def run_fix(self):
        execute_command(["systemctl", "enable", self.SERVICE_NAME], check_returncode=True)

    def get_description(self):
        return "Checking that {} exists and enabled".format(self.DST_FILE)

    def get_kb_number(self):
        return "202696031"  # TODO sync my kb article to support.infinidat.com, copy the number, create a new qr code
