
from ....case import Case, Category
from infi.pyutils.lazy import cached_method
from infi.winver import Windows
from infi.registry import LocalComputer
from infi.winapi.file_version_information import File
from infi.os_info.parse_version import parse_version
import os
import glob

from logging import getLogger
log = getLogger(__name__)

HOTFIX_BASE = 'SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\Hotfix'
HOTFIX_CACHE_DIR = os.path.join(os.environ.get('SystemRoot', r"C:\Windows"), 'winsxs')

class HotfixTestCase(Case):
    def _get_windows_kb_number(self):
        return self._get_windows_kb_dict()['kb']

    def get_description(self):
        msg = "Checking that driver {} is up-to-date with required KB fixes"
        return msg.format(self.component_name)

    def get_kb_number(self):
        return '202404321'

    @cached_method
    def _get_windows_version(self):
        return Windows().version

    @cached_method
    def _get_windows_service_pack_tag(self):
        number = Windows().service_pack
        return "SP{}".format(number) if number else "RTM"

    def _get_case_config(self):
        from infi.vendata.powertools.conf import get_configuration
        return get_configuration().get('readiness.cases.recommended_hotfixes')

    def _get_windows_kb_dict(self):
        component_dict = self._get_case_config()[self.component_name]
        return component_dict[self._get_windows_version()][self._get_windows_service_pack_tag()]

    def _get_expected_hotfix_version(self):
        version = self._get_windows_kb_dict()['version']
        log.debug("expected version of {} is {!r}".format(self.component_name, version))
        return version

    def _is_windows_kb_installed(self):
        try:
            key = LocalComputer().local_machine[HOTFIX_BASE]['KB{}'.format(self._get_windows_kb_number())]
            value = key.values_store['Installed'].to_python_object()
            return value == 1
        except KeyError:
            pass
        return False

    def _is_newer_hotfix_installed(self):
        return self._is_install_component_file_version_newer_than_expected()

    def _get_version_of_file(self, filepath):
        from os.path import exists
        if not exists(filepath):
            self.skip("cannot determine if hotfix is installed or not since its feature is not installed")
        version = File(filepath).get_version()
        log.debug("actual version of {} is {!r}".format(filepath, version))
        return version

    def _get_component_filename(self):
        return os.path.join(os.environ.get("SystemRoot", os.path.join("C:", "Windows")),
                            'System32', 'Drivers', self.component_name)

    def _get_component_version(self):
        return self._get_version_of_file(self._get_component_filename())

    def _is_install_component_file_version_newer_than_expected(self):
        actual = self._get_component_version()
        expected = self._get_expected_hotfix_version()
        return parse_version(actual) >= parse_version(expected)

    def _iter_cached_files_for_component(self):
        pathname = os.path.join(HOTFIX_CACHE_DIR, '*{}*'.format(self.component_name.split('.')[0]))
        for directory in glob.glob(pathname):
            possible_file = os.path.join(HOTFIX_CACHE_DIR, directory, self.component_name)
            if os.path.isfile(possible_file):
                yield possible_file

    def _is_reboot_required(self):
        expected = self._get_expected_hotfix_version()
        for cached_component_file in self._iter_cached_files_for_component():
            actual = self._get_version_of_file(cached_component_file)
            if parse_version(actual) >= parse_version(expected):
                return True
        return False

    def run_test(self):
        try:
            self._get_windows_kb_dict()
        except ValueError:
            self.warn("Unknown operating system {} {}".format(self._get_windows_version(), self._get_windows_service_pack_tag()))
            return

        if self._is_windows_kb_installed():
            pass
        elif self._is_newer_hotfix_installed():
            pass
        elif self._is_reboot_required():
            msg = "There is an update pending for {}"
            self.ask_for_reboot(msg.format(self.component_name))
        else:
            msg = "Windows KB{} (or a newer hotfix) is not installed for {}"
            self.warn(msg.format(self._get_windows_kb_number(), self.component_name))

class MPIOSYS(HotfixTestCase):
    component_name = 'mpio.sys'

class MSDSMSYS(HotfixTestCase):
    component_name = 'msdsm.sys'

def get_category():
    return Category("Hotfixes", [MPIOSYS, MSDSMSYS])
