import re
import os
import glob

DEFAULT_GRUB_FILE = os.path.join(os.path.sep, 'etc', 'grub.conf')
DEFAULT_GRUB2_FILE = (glob.glob(os.path.join(os.path.sep, 'boot', 'grub*', 'grub.cfg')) or ['/path/does/not/exist'])[0]
BOOT_DIRPATH = os.path.join(os.path.sep, 'boot')


class Grub1BootEntry(object):
    TITLE_PATTERN = r"^title (?P<title>.*)"
    ROOT_DEVICE_PATTERN = r"^[\t ]*root (?P<root_device>.*)"
    KERNEL_PATTERN = r"^[\t ]*kernel (?P<kernel_image>[^ ]*) (?P<kernel_parameters>.*)"
    INITRD_PATTERN = r"(^[\t ]*)initrd (?P<initrd_image>.*)"

    def __init__(self, string):
        self._string = string

    def _return_match_for_pattern(self, pattern, key):
        matchobj = re.search(pattern, self._string, re.MULTILINE)
        return matchobj if matchobj is None else matchobj.groupdict()[key]

    def get_title(self):
        return self._return_match_for_pattern(self.TITLE_PATTERN, "title")

    def get_root_device(self):
        return self._return_match_for_pattern(self.ROOT_DEVICE_PATTERN, "root_device")

    def get_kernel_image(self):
        return self._return_match_for_pattern(self.KERNEL_PATTERN, "kernel_image")

    def get_kernel_parameters(self):
        return self._return_match_for_pattern(self.KERNEL_PATTERN, "kernel_parameters")

    def _get_initrd_image(self):
        return self._return_match_for_pattern(self.INITRD_PATTERN, "initrd_image")

    def get_initrd_image(self):
        location = self._get_initrd_image()
        if os.path.ismount(BOOT_DIRPATH):
            return os.path.join(BOOT_DIRPATH, location.strip(os.path.sep))
        return location

    def set_title(self, title):
        self._string = self._string.replace(self.get_title(), title)

    def set_initrd_image(self, initrd_image):
        if os.path.ismount(BOOT_DIRPATH):
            initrd_image = initrd_image.replace(BOOT_DIRPATH, '')
        self._string = self._string.replace(self._get_initrd_image(), initrd_image)

    @classmethod
    def from_grub_conf(cls, string):
        return cls(string)

    def __eq__(self, other):
        keys = ['get_title', 'get_root_device', 'get_kernel_image', 'get_kernel_parameters', 'get_initrd_image']
        return any([getattr(self, key)() == getattr(other, key)() for key in keys])

    def __repr__(self):
        return "grub.conf boot entry <{}>".format(self.get_title())

    def __str__(self):
        return self._string


class Grub2BootEntry(Grub1BootEntry):
    MENUENTRY_PATTERN = "menuentry (['\"])(.*?)(['\"])"
    INITRD_PATTERN = r"\tinitrd\d*\s+(.*)\n"

    def __init__(self, string):
        self._string = string

    def get_title(self):
        return re.match(self.MENUENTRY_PATTERN, self._string).group(2)

    def get_initrd_image(self):
        matchobj = re.search(self.INITRD_PATTERN, self._string)
        if matchobj is None:
            return None
        if os.path.ismount(BOOT_DIRPATH):
            return os.path.join(BOOT_DIRPATH, matchobj.group(1).strip(os.path.sep))
        return matchobj.group(1)

    def set_title(self, title):
        self._string = re.sub(self.MENUENTRY_PATTERN, r"menuentry \1{}\3".format(title), self._string)

    def set_initrd_image(self, initrd_image):
        self._string = re.sub(self.INITRD_PATTERN, "\tinitrd\t{}\n".format(initrd_image), self._string)

    @classmethod
    def from_grub_conf(cls, string):
        # it's not trivial to parse Grub 2 boot entries, so we just keep the entire string and manipulate it
        return cls(string)

    def __repr__(self):
        return "Grub 2 boot entry <{}>".format(self.get_title())

    def __str__(self):
        return self._string


class Grub1Config(object):
    def __init__(self, filepath=DEFAULT_GRUB_FILE):
        super(Grub1Config, self).__init__()
        self._filepath = filepath

    def _read_file(self):
        with open(self._filepath) as fd:
            return fd.read()

    def get_boot_entries(self):
        pattern = re.compile(r"^title.*\n(?:(?!title)[\t ]*\S+.*\n)*", re.MULTILINE)
        return [Grub1BootEntry.from_grub_conf(match) for match in re.findall(pattern, self._read_file())]

    def add_boot_entry(self, boot_entry):
        with open(self._filepath, 'a') as fd:
            fd.write("{}".format(str(boot_entry)))

    def create_backup(self):
        import tempfile
        import shutil
        basedir = os.path.dirname(self._filepath)
        filename = os.path.basename(self._filepath)
        fd, path = tempfile.mkstemp(prefix=filename, dir=basedir)
        os.close(fd)
        shutil.copyfile(self._filepath, path)
        return path

    def remove_boot_entry(self, entry):
        original_content = self._read_file()
        modified_content = original_content.replace(str(entry), '')
        with open(self._filepath, 'w') as fd:
            fd.write(modified_content)

    def _get_default_boot_entry_number(self):
        pattern = re.compile("^default[=\s](?P<index>\d+$)", re.MULTILINE)
        return int(re.search(pattern, self._read_file()).groupdict()['index'])

    def get_default_boot_entry(self):
        boot_entries = self.get_boot_entries()
        index = self._get_default_boot_entry_number()
        return boot_entries[index]


class Grub2Config(Grub1Config):
    def __init__(self, filepath=DEFAULT_GRUB2_FILE):
        super(Grub2Config, self).__init__(filepath)

    def get_boot_entries(self):
        data = self._read_file()
        entry_texts = re.findall("menuentry.*?{\n(?:\t.*?(?: .*?)?\n)*}", data)
        return [Grub2BootEntry.from_grub_conf(entry_text) for entry_text in entry_texts]

    def _get_default_boot_entry_number(self):
        default_data = open("/etc/default/grub", "rb").read()
        grub_default = re.search("GRUB_DEFAULT=(\d+)", default_data)
        if grub_default is None:
            return 0
        return int(grub_default.group(1))

    def add_boot_entry(self, boot_entry):
        DEFAULT_CUSTOM_MENU_PATH = os.path.join(os.path.sep, "etc", "grub.d", "40_custom")
        with open(DEFAULT_CUSTOM_MENU_PATH, 'a+') as fd:
            fd.write("\n")
            fd.write(str(boot_entry))

    def _update_grub(self):
        from ....utils import execute_command
        if os.path.exists("/usr/sbin/update-grub"):
            execute_command(["/usr/sbin/update-grub"])
        elif os.path.exists("/usr/sbin/grub2-mkconfig"):
            execute_command(["/usr/sbin/grub2-mkconfig", "-o", "/boot/grub2/grub.cfg"])
        else:
            raise NotImplementedError()

    def remove_boot_entry(self, entry):
        raise NotImplementedError()


if os.path.exists(DEFAULT_GRUB2_FILE):
    GrubConfig = Grub2Config
    GrubBootEntry = Grub2BootEntry
else:
    GrubConfig = Grub1Config
    GrubBootEntry = Grub1BootEntry
