
import sys

from infi.vendata.powertools.utils import queries
from . import actions
from logging import getLogger
from infi.exceptools import InfiException
from infi.pyutils.contexts import contextmanager

log = getLogger(__name__)

output_stream = sys.stdout
DEFAULT_PROMPT = "Are you sure?"

class EngineException(InfiException):
    pass


class Abort(EngineException):
    pass


# Convert list of infinisdk lun objects to dict
def lun_dict(luns):
    return {int(lun): lun.volume for lun in luns}


class HostRegistration(object):
    # pylint: disable=R0902,R0913,W0201

    def __init__(self, vendor_mixin, connections, host_name, fix, automatic_fix):
        object.__init__(self)
        self.vendor_mixin = vendor_mixin
        self.host_name = host_name
        self.connections = connections
        self.fix = fix
        self.automatic_fix = automatic_fix
        self._populate_objects()

    def _populate_objects(self):
        self.system_object = queries.get_system_object_by_device_vendor(self.vendor_mixin)
        self.our_ports = list({item.get_initiator_wwn() for item in self.connections})

    def get_available_host_name(self):
        """ Find a name of a host that doesn't exist or exists but doesn't have foreign ports in it """
        host_name = self.host_name
        i = 1
        while True:
            if not queries.is_host_object_exists(self.system_object, host_name):
                break
            host_object = queries.get_host_object_by_name(self.system_object, host_name)
            if not queries.are_there_foreign_ports_on_host(host_object, self.our_ports):
                break
            host_name = "{}-{}".format(self.host_name, i)
            i += 1
        return host_name

    def _cleanup_host_object_from_foreign_ports(self, host_object):
        msg = "Host name {} contains fc ports that are not physically on this server: {!r}"
        foreign_ports = queries.get_foreign_ports_from_host(host_object, self.our_ports)
        self.ask_user(msg.format(host_object.get_name(), foreign_ports), "Remove these ports")
        for foreign_port in foreign_ports:
            with self._temporarily_remove_luns_from_host(foreign_port, host_object, lun_dict(host_object.get_luns()).values()):
                actions.remove_port_from_host(host_object, foreign_port)

    def _choose_another_hostname(self):
        suggested_host_name = self.get_available_host_name()
        msg = "Host name {} contains fc ports that are not physically on this server"
        self.ask_user(msg.format(self.host_name), "Register under a different name?")
        notify_user("Registering under the name {}".format(suggested_host_name))
        self.host_name = suggested_host_name

    def _rename_host_if_necessary(self):
        host_object = queries.get_host_object_by_wwn(self.system_object, self.our_ports[0])
        if host_object.get_name() == self.host_name:
            msg = "Initiator ports {!r} is already registered in host {} on system {} ({})"
            notify_user(msg.format(self.our_ports, self.host_name,
                                   self.system_object.get_serial(), self.system_object.get_name()))
        else:
            self.ask_user("Host name does not match ({} != {})".format(host_object.get_name(), self.host_name),
                      "Rename host?")
            msg = "Host has been renamed from {} to {}"
            previous_name = host_object.get_name()
            host_object.update_field('name', self.host_name)
            notify_user(msg.format(previous_name, self.host_name))

    def _merge_lun_mappings(self, host_object, other_host):
        # cluster membership
        other_cluster = other_host.get_cluster()
        cluster_object = other_host.get_cluster()
        our_luns = lun_dict(cluster_object.get_luns()) if cluster_object else {}
        our_luns.update(lun_dict(host_object.get_luns()))

        lun_mappings_to_migrate = dict()
        if other_cluster:
            if cluster_object is None:
                # our host not a member of any cluster, we can just add our host
                other_cluster.add_host(host_object)
                msg = "Host {} added as a member to cluster {}"
                notify_user(msg.format(host_object.get_name(), other_cluster.get_name()))
            else:
                for lun, volume in lun_dict(other_cluster.get_luns()).items():
                    if volume in our_luns.values():
                        continue
                    lun_mappings_to_migrate[lun] = volume
        for lun, volume in lun_dict(other_host.get_luns()).items():
            if volume in our_luns.values():
                continue
            lun_mappings_to_migrate[lun] = volume

        for lun, volume in lun_mappings_to_migrate.items():
            if lun in our_luns:
                msg = "Volume {} is already mapped to LUN {!r}"
                self.ask_user(msg.format(volume.get_name(), lun), "Map to a diffrent LUN?")
                host_object.map_volume(volume)
            else:
                host_object.map_volume(volume, lun)
            msg = "Volume {} mapped to host {}"
            notify_user(msg.format(volume.get_name(), host_object.get_name()))
        return lun_mappings_to_migrate

    def _rescan_for_volumes(self, volumes=[]):
        from infi.storagemodel import get_storage_model
        from infi.storagemodel.predicates import PredicateList
        from infi.storagemodel.vendor.infinidat.predicates import InfinidatVolumeExists
        predicates = [InfinidatVolumeExists(self.system_object.get_serial(), volume.get_id()) for
                      volume in volumes]
        get_storage_model().rescan_and_wait_for(PredicateList(predicates))

    @contextmanager
    def _temporarily_remove_luns_from_host(self, fc_port, other_host, volumes):
        luns = {key: value for key, value in lun_dict(other_host.get_luns()).items() if value in volumes}
        msg = "Before moving the initiator port {!r} from host {}, its volumes must be temporarily unmapped"
        self.ask_user(msg.format(fc_port, other_host.get_name()), 'Unmap the volumes?')
        for volume in luns.values():
            other_host.unmap_volume(volume)
            msg = "Volume {} temporarily unmapped from host {}"
            notify_user(msg.format(volume.get_name(), other_host.get_name()))
        try:
            yield
        finally:
            for lun, volume in luns.items():
                other_host.map_volume(volume, lun)
                msg = "Volume {} mapped back to host {}"
                notify_user(msg.format(volume.get_name(), other_host.get_name()))

    def register_all_connections(self):
        expected_lun_mappings = {}
        if queries.is_host_object_exists(self.system_object, self.host_name):
            log.info("host object exists")
            host_object = queries.get_host_object_by_name(self.system_object, self.host_name)
            self.host_object = host_object
            if queries.are_there_foreign_ports_on_host(host_object, self.our_ports):
                log.info("host object exists, and it has foreign ports")
                if set(host_object.get_fc_ports()).intersection(self.our_ports):
                    # our ports are partially registered in another host, which also has other, unidentified fc ports
                    # this can happen when the user replaced fc cards
                    log.info("host object exists, and it has only foreign ports")
                    self._cleanup_host_object_from_foreign_ports(host_object)
                else:
                    # neither of our ports is registered under the requested host name
                    # so we have a name conflict
                    log.info("host object exists, and it has only foreign ports in addition to our ports")
                    self._choose_another_hostname()
        if not queries.is_any_of_our_ports_registered_in_system(self.system_object, self.our_ports):
            log.info("at least one of our ports is registered in the box")
            if not queries.is_host_object_exists(self.system_object, self.host_name):
                self.host_object = actions.create_host(self.system_object, self.host_name)
                notify_user("Created host {}".format(self.host_name))
            host_object = queries.get_host_object_by_name(self.system_object, self.host_name)
            actions.add_ports_to_host(host_object, self.our_ports)
            if 0 != len(self.our_ports):
                msg = "Registered initiator ports {!r} in host {} on system {} ({})"
                notify_user(msg.format(self.our_ports, self.host_name, self.system_object.get_serial(), self.system_object.get_name()))
            else:
                msg = "Could not register any initiator ports in host {} on system {} ({}). Please contact support"
                notify_user(msg.format(self.host_name, self.system_object.get_serial(), self.system_object.get_name()))
            self.host_object = host_object
        elif queries.are_all_ports_belong_to_the_same_host(self.system_object, self.our_ports):
            self.host_object = queries.get_host_objects_by_our_ports(self.system_object, self.our_ports)[0]
            log.info("all of our ports are registered under the same host without any other port")
            self._rename_host_if_necessary()
        else:
            # baglan, some of our ports are registered, some are not
            log.info("balagan, some of our ports are already registered, not necessarily under the same host")
            if not queries.is_host_object_exists(self.system_object, self.host_name):
                actions.create_host(self.system_object, self.host_name)
                notify_user("Created host {}".format(self.host_name))
            host_object = queries.get_host_object_by_name(self.system_object, self.host_name)
            pending_ports = set(self.our_ports)

            # first, lets deal with the simple cases
            for our_port in list(pending_ports):
                other_host = queries.get_host_object_by_wwn(self.system_object, our_port)
                if other_host is None:
                    # port is not registered, lets just register it
                    actions.add_ports_to_host(host_object, [our_port])
                    msg = "Registered initiator port {!r} in host {} on system {} ({})"
                    notify_user(msg.format(our_port, self.host_name, self.system_object.get_serial(), self.system_object.get_name()))
                    self.host_object = host_object
                    pending_ports.discard(our_port)
                elif other_host.get_id() == host_object.get_id():
                    # port is registered under the right host, nothing to do here
                    msg = "Initiator port {!r} is already registered in host {} on system {} ({})"
                    notify_user(msg.format(our_port, self.host_name, self.system_object.get_serial(), self.system_object.get_name()))
                    pending_ports.discard(our_port)
            # now, we need to to deal with the problematic ports
            for our_port in pending_ports:
                # balagan, the port is registered under a different host
                # first, we need to copy the lun mappings
                # then, move the port and rescan
                other_host = queries.get_host_object_by_wwn(self.system_object, our_port)
                log.debug("port is mapped to a different host, we need to move it")
                msg = "Initiator port {!r} is registered under a different host ({})"
                self.ask_user(msg.format(our_port, other_host.get_name()), "Move LUN mappings and port?")
                expected_lun_mappings.update(self._merge_lun_mappings(host_object, other_host))
                self._rescan_for_volumes(expected_lun_mappings.values())
                with self._temporarily_remove_luns_from_host(our_port, other_host, expected_lun_mappings.values()):
                    actions.remove_port_from_host(other_host, our_port)
                    notify_user("Removed port {!r} from host {}".format(our_port, other_host.get_name()))
                actions.add_ports_to_host(host_object, [our_port])
                msg = "Registered initiator port {!r} in host {} on system {} ({})"
                notify_user(msg.format(our_port, self.host_name, self.system_object.get_serial(), self.system_object.get_name()))
                self.host_object = host_object

                if queries.are_there_no_ports_on_host(other_host):
                    msg = "Host {} is now empty (no ports defined)"
                    self.ask_user(msg.format(other_host.get_name()), "Delete host?")
                    other_host_name = other_host.get_name()
                    actions.delete_host_from_box(other_host)
                    notify_user("Deleted host {}".format(other_host_name))

        self._rescan_for_volumes(expected_lun_mappings.values())

    def ask_user(self, problem_description, prompt=DEFAULT_PROMPT):
        print "WARNING: {}".format(problem_description)
        if not self.fix and not self.automatic_fix:
            if prompt == DEFAULT_PROMPT:
                print "In order to fix it, use --fix or --auto-fix flags"
            else:
                print "In order to fix it ({!r}), use --fix or --auto-fix flags".format((prompt[0].lower() + prompt[1:]).strip('?'))
            raise Abort(problem_description)
        if not self.automatic_fix:
            if not confirm(prompt):
                print "Aborted by user request"
                raise Abort(problem_description)


def confirm(prompt=None, resp=False):  # pragma: no cover
    """prompts for yes or no response from the user. Returns True for yes and
    False for no.

    'resp' should be set to the default value assumed by the caller when
    user simply types ENTER.

    >>> confirm(prompt='Create Directory?', resp=True)
    Create Directory? [y]|n:
    True
    >>> confirm(prompt='Create Directory?', resp=False)
    Create Directory? [n]|y:
    False
    >>> confirm(prompt='Create Directory?', resp=False)
    Create Directory? [n]|y: y
    True

    """

    if prompt is None:
        prompt = 'Confirm'

    if resp:
        prompt = '%s [%s]|%s: ' % (prompt, 'y', 'n')
    else:
        prompt = '%s [%s]|%s: ' % (prompt, 'n', 'y')

    while True:
        ans = raw_input(prompt)
        if not ans:
            return resp
        if ans not in ['y', 'Y', 'n', 'N']:
            print 'please enter y or n.'
            continue
        if ans == 'y' or ans == 'Y':
            return True
        if ans == 'n' or ans == 'N':
            return False


def notify_user(message):
    log.info("Notification to user: {}".format(message))
    print >>output_stream, message


def update_metadata(host, **kwargs):
    from infinisdk.core.exceptions import APICommandFailed
    try:
        metadata = host.get_all_metadata()
        changed = False
        for key, value in kwargs.iteritems():
            message = "    {} = {!r}".format(key, value)
            notify_user(message)
            if metadata.get(key) != value:
                changed = True
        if changed:
            host.set_metadata_from_dict(kwargs)
    except APICommandFailed as error:
        if error.status_code == 403:
            log.warn(str(error))
        else:
            raise


def update_metadata_on_host_object(host_object):
    update_metadata(host_object,
                    hostname=queries.get_os_hostname(),
                    system=queries.get_os_system(),
                    platform=queries.get_os_platform(),
                    powertools_version=queries.get_powertools_version())


def update_metadata_on_controller(controller):
    system_object = queries.get_system_object_by_device(controller)
    if not queries.is_host_port_defined(controller):
        raise Abort("This host is not properly defined on system {!r}".format(system_object.get_serial()))
    msg = "Updated metadata in system {!r} for host id {!r}"
    host_object = queries.get_defined_host_object(system_object, controller)
    notify_user(msg.format(system_object.get_serial(), host_object.get_id()))
    update_metadata_on_host_object(host_object)


@contextmanager
def output_stream_context(stream=None):
    global output_stream
    old_stream = output_stream
    output_stream = old_stream if stream is None else stream
    try:
        yield
    finally:
        output_stream = old_stream


def register_host_in_box(namespace):
    boxes = {key: value for
             key, value in queries.get_connections_by_serial().items() if
             key == int(namespace["--system"])}
    host_name = namespace["--hostname"]
    if not boxes:
        notify_user("No initiator WWNs found for InfiniBox {}, please contact support".format(namespace["--system"]))
        return
    [(vendor_mixin, connections)] = boxes.values()
    registration = HostRegistration(vendor_mixin, connections, host_name, namespace["--fix"], namespace["--auto-fix"])
    registration.register_all_connections()
    return registration


def register(namespace, stream=None):
    from ...rescan.engine import clear_unit_attention_from_infinidat_storage_controllers
    with output_stream_context(stream):
        registration = register_host_in_box(namespace)
        update_metadata_on_host_object(registration.host_object)
        clear_unit_attention_from_infinidat_storage_controllers()


def in_register(arguments):
    from ...rescan.engine import clear_unit_attention_from_infinidat_storage_controllers
    try:
        register(arguments)
        clear_unit_attention_from_infinidat_storage_controllers()
        return 0
    except Abort:
        return 1
