
from infi.unittest import TestCase, SkipTest
from infi.unittest.parameters import iterate
from mock import patch

from .. import HostRegistration, Abort
from infi.credentials_store import Credentials

# pylint: disable-all

from infi.vendata.smock import HostMock
from infi.run_as_root import RootPermissions
from infi.vendata.smock.inventory import generate_wwn
from infi.vendata.powertools.system_register.engine import lun_dict

class RegistrationTestCase(TestCase):
    def setUp(self):
        super(RegistrationTestCase, self).setUp()
        from infi.vendata.powertools.utils import auth
        from infi.vendata.powertools.utils.queries import get_connections_by_serial
        auth.CREDENTIALS_STORE.set_default_credentials(Credentials('admin', '123456'))

        self.is_root = patch.object(RootPermissions, "is_root")
        self.smock = HostMock()
        self.inventory = self.smock.get_inventory()
        self.system = self.inventory.add_infinibox()
        self.initiators = [self.inventory.add_initiator(), ]
        self.inventory.zone_with_system__best_practices(self.system)
        self.hostname = self.inventory.get_hostname()
        self.smock.start()
        self.connections = get_connections_by_serial().values()[0][-1]
        self.is_root.start()
        self.smock._clear_storagemodel_cache()

    def tearDown(self):
        self.smock.end()
        self.is_root.stop()
        super(RegistrationTestCase, self).tearDown()

    def teardown_simulator(self):
        self.simulator_process.terminate()

    def create_test_host(self):
        self.host = self.inventory.register_host_in_system(self.system)
        return self.host

    def assert_host_exists(self, host_name=None, port_address=None):
        matching_hosts = [host for host in self.system.hosts if host.get_name() == (host_name or self.hostname)]
        self.assertEqual(len(matching_hosts), 1, "No matching host found")
        self.assertEqual(set(matching_hosts[0].get_fc_ports()), set([(port_address or self.initiators[0])]))

    def get_scsi_controller(self):
        from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_storage_controller_devices
        return get_infinidat_storage_controller_devices()[0]

    def test_already_defined(self):
        self.create_test_host()
        reg = HostRegistration(self.get_scsi_controller().get_vendor(), self.connections, self.hostname, False, False)
        reg.register_all_connections()
        self.assert_host_exists()

    def test_host_does_not_exist(self):
        self.assertEqual(self.system.hosts.count(), 0)
        reg = HostRegistration(self.get_scsi_controller().get_vendor(), self.connections, self.hostname, False, False)
        reg.register_all_connections()
        self.assert_host_exists()

    @iterate("fix", [True, False])
    @iterate("auto_fix", [True, False])
    def test_host_name_mismatch__no_volumes_mapped__no_other_ports(self, fix, auto_fix):
        if not fix and auto_fix:
            self.skipTest("invalid args")
        self.create_test_host()
        reg = HostRegistration(self.get_scsi_controller().get_vendor(), self.connections, "otherName", fix, auto_fix)
        if fix and auto_fix:
            reg.register_all_connections()
            self.assert_host_exists("otherName")
        elif fix and not auto_fix:
            with patch("infi.vendata.powertools.system_register.engine.confirm") as confirm_patch:
                confirm_patch.return_value = auto_fix
                self.assertRaises(Abort, reg.register_all_connections)
        else:
            self.assertRaises(Abort, reg.register_all_connections)

    @iterate("fix", [True, False])
    @iterate("auto_fix", [True, False])
    def test_host_name_mismatch__no_volumes_mapped__other_ports_exist(self, fix, auto_fix):
        if not fix and auto_fix:
            self.skipTest("invalid args")
        host = self.create_test_host()
        host.add_fc_port(generate_wwn())
        reg = HostRegistration(self.get_scsi_controller().get_vendor(), self.connections, "otherName", fix, auto_fix)
        if fix and auto_fix:
            reg.register_all_connections()
            return
        elif fix and not auto_fix:
            with patch("infi.vendata.powertools.system_register.engine.confirm") as confirm_patch:
                confirm_patch.return_value = auto_fix
                self.assertRaises(Abort, reg.register_all_connections)
        else:
            self.assertRaises(Abort, reg.register_all_connections)

    @iterate("fix", [True, False])
    @iterate("auto_fix", [True, False])
    def test_host_name_mismatch__volumes_mapped(self, fix, auto_fix):
        # Register host as name A with port W1 when already exists: host B with ports W1 and Q1
        # B should stay with Q1 only, A should be created with W1 and the volumes should be mapped to both
        # A="otherName", B=self.hostname
        from capacity import GB, TiB
        if not fix and auto_fix:
            self.skipTest("invalid args")
        foreign_port = generate_wwn()
        host = self.create_test_host()
        host.add_fc_port(foreign_port)
        pool = self.system.pools.create(name="pool", virtual_capacity=3 * TiB, physical_capacity=3 * TiB)
        volume = self.system.volumes.create(pool=pool, name="volume1", size=1 * GB, provisioning="THICK")
        host.map_volume(volume)
        self.smock._clear_storagemodel_cache()
        reg = HostRegistration(self.get_scsi_controller().get_vendor(), self.connections, "otherName", fix, auto_fix)
        if fix and auto_fix:
            reg.register_all_connections()
            self.assert_host_exists(self.hostname, foreign_port)
            self.assert_host_exists("otherName", self.initiators[0])
            new_host = [host for host in self.system.hosts if host.get_name() == "otherName"][0]
            self.assertIn(volume, lun_dict(new_host.get_luns()).values())
        elif fix and not auto_fix:
            with patch("infi.vendata.powertools.system_register.engine.confirm") as confirm_patch:
                confirm_patch.return_value = auto_fix
                self.assertRaises(Abort, reg.register_all_connections)

    @iterate("fix", [True, False])
    @iterate("auto_fix", [True, False])
    def test_port_doesnt_exist_in_host__with_foreign_ports(self, fix, auto_fix):
        if not fix and auto_fix:
            self.skipTest("invalid args")
        foreign_wwn = generate_wwn()
        host = self.create_test_host()
        host.remove_fc_port(self.initiators[0])
        host.add_fc_port(foreign_wwn)
        reg = HostRegistration(self.get_scsi_controller().get_vendor(), self.connections, self.hostname, fix, auto_fix)
        if fix and auto_fix:
            reg.register_all_connections()
            self.assert_host_exists(self.hostname + "-1", self.initiators[0])
            self.assert_host_exists(self.hostname, foreign_wwn)
        elif fix and not auto_fix:
            with patch("infi.vendata.powertools.system_register.engine.confirm") as confirm_patch:
                confirm_patch.return_value = auto_fix
                self.assertRaises(Abort, reg.register_all_connections)

    @iterate("fix", [True, False])
    @iterate("auto_fix", [True, False])
    def test_port_doesnt_exist_in_host__no_foreign_ports(self, fix, auto_fix):
        host = self.create_test_host()
        host.remove_fc_port(self.initiators[0])
        reg = HostRegistration(self.get_scsi_controller().get_vendor(), self.connections, self.hostname, fix, auto_fix)
        reg.register_all_connections()
        self.assert_host_exists()

    @iterate("fix", [True, False])
    @iterate("auto_fix", [True, False])
    def test_is_registered_with_proper_name__with_foreign_ports(self, fix, auto_fix):
        if not fix and auto_fix:
            self.skipTest("invalid args")
        foreign_port = generate_wwn()
        host = self.create_test_host()
        host.add_fc_port(foreign_port)
        reg = HostRegistration(self.get_scsi_controller().get_vendor(), self.connections, self.hostname, fix, auto_fix)
        if fix and auto_fix:
            reg.register_all_connections()
            self.assert_host_exists(self.hostname, self.initiators[0])
        elif fix and not auto_fix:
            with patch("infi.vendata.powertools.system_register.engine.confirm") as confirm_patch:
                confirm_patch.return_value = auto_fix
                self.assertRaises(Abort, reg.register_all_connections)

    def test_is_host_port_defined__port_defined(self):
        from ....utils.queries import is_host_port_defined
        host = self.create_test_host()
        host.add_fc_port(generate_wwn())
        self.assertTrue(is_host_port_defined(self.get_scsi_controller()))

    def test_is_host_port_defined__port_not_defined(self):
        from ....utils.queries import is_host_port_defined
        self.assertFalse(is_host_port_defined(self.get_scsi_controller()))

class RegistrationTestCase2(TestCase):
    def setUp(self):
        super(RegistrationTestCase2, self).setUp()

        from infi.vendata.powertools.utils import auth
        auth.CREDENTIALS_STORE.set_default_credentials(Credentials('admin', '123456'))
        self.smock = HostMock()
        self.inventory = self.smock.get_inventory()
        self.system = self.inventory.add_infinibox()
        self.our_port1 = self.inventory.add_initiator()
        self.our_port2 = self.inventory.add_initiator()
        self.our_ports = [self.our_port1, self.our_port2]
        self.inventory.zone_with_system__best_practices(self.system)
        self.hostname = self.inventory.get_hostname()

        self.other_port1 = generate_wwn()
        self.other_port2 = generate_wwn()
        self.other_port3 = generate_wwn()
        self.smock.start()
        self.smock._clear_storagemodel_cache()

    def tearDown(self):
        self.smock.end()
        super(RegistrationTestCase2, self).tearDown()

    def assert_host_exists(self, host_name, ports):
        matching_hosts = [host for host in self.system.hosts if host.get_name() == host_name]
        self.assertEqual(len(matching_hosts), 1, "No matching host found")
        self.assertEqual(set(matching_hosts[0].get_fc_ports()), set(ports))

    def assert_host_deleted(self, host_name):
        host_names = [host.get_name() for host in self.system.hosts]
        self.assertNotIn(host_name, host_names)

    def run_register(self):
        from .. import register_host_in_box
        namespace = {"--system": self.system.get_serial(),
                     "--hostname": self.hostname,
                     "--auto-fix": True,
                     "--fix": False}
        register_host_in_box(namespace)

    def register_host(self, host_name, host_ports):
        host = self.system.hosts.create(name=host_name)
        for port in host_ports:
            host.add_fc_port(port)
        return host

    def get_scsi_controller(self):
        from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_storage_controller_devices
        return get_infinidat_storage_controller_devices()[0]

    def create_volume(self):
        from capacity import GB, TiB
        pool = self.system.pools.create(name="pool", virtual_capacity=3 * TiB, physical_capacity=3 * TiB)
        volume = self.system.volumes.create(pool=pool, name="volume1", size=1 * GB, provisioning="THICK")
        self.smock._clear_storagemodel_cache()
        return volume

    def test_normal(self):
        """ Register Host A with W1 and W2 when Host A is not yet registered """
        self.run_register()
        self.assert_host_exists(self.hostname, self.our_ports)

    def test_host_exists(self):
        """ Register Host A with W1 and W2 when Host A with no ports is already registered """
        self.register_host(self.hostname, [])
        self.run_register()
        self.assert_host_exists(self.hostname, self.our_ports)

    def test_already_registered(self):
        """ Register Host A with W1 and W2 when Host A with W1 and W2 is already registered """
        self.register_host(self.hostname, [self.our_port1, self.our_port2])
        self.run_register()
        self.assert_host_exists(self.hostname, self.our_ports)

    def test_registered_different_ports(self):
        """ Register Host A with W1 and W2 when Host A with Q1 and Q2 is already registered """
        self.register_host(self.hostname, [self.other_port1, self.other_port2])
        self.run_register()
        self.assert_host_exists(self.hostname, [self.other_port1, self.other_port2])
        self.assert_host_exists(self.hostname + "-1", self.our_ports)

    def test_registered_mixed(self):
        """ Register Host A with W1 and W2 when Host A with W1 and Q1 is already registered """
        self.register_host(self.hostname, [self.our_port1, self.other_port1])
        self.run_register()
        self.assert_host_exists(self.hostname, self.our_ports)

    def test_one_port_registered(self):
        """ Register Host A with W1 and W2 when Host A with W1 already registered """
        self.register_host(self.hostname, [self.our_port1])
        self.run_register()
        self.assert_host_exists(self.hostname, self.our_ports)

    def test_registered_with_extra_port(self):
        """ Register Host A with W1 and W2 when Host A with W1, W2 and Q1 already registered """
        self.register_host(self.hostname, [self.our_port1, self.our_port2, self.other_port1])
        self.run_register()
        self.assert_host_exists(self.hostname, self.our_ports)

    def test_registered_as_different_host(self):
        """ Register Host A with W1 and W2 when Host B with W1 and W2 already registered """
        self.register_host("foo", [self.our_port1, self.our_port2])
        self.run_register()
        self.assert_host_deleted("foo")
        self.assert_host_exists(self.hostname, self.our_ports)

    def test_one_port_in_different_host_mixed(self):
        """ Register Host A with W1 and W2 when Host B with W1 and Q1 already registered """
        self.register_host("foo", [self.our_port1, self.other_port1])
        self.run_register()
        self.assert_host_exists("foo", [self.other_port1])
        self.assert_host_exists(self.hostname, self.our_ports)

    def test_one_port_in_different_host(self):
        """ Register Host A with W1 and W2 when Host B with W1 already registered """
        self.register_host("foo", [self.our_port1])
        self.run_register()
        self.assert_host_deleted("foo")
        self.assert_host_exists(self.hostname, self.our_ports)

    def test_registered_as_different_host_with_extra_port(self):
        """ Register Host A with W1 and W2 when Host B with W1, W2 and Q1 already registered """
        self.register_host("foo", [self.our_port1, self.our_port2, self.other_port1])
        self.run_register()
        self.assert_host_exists("foo", [self.other_port1])
        self.assert_host_exists(self.hostname, self.our_ports)

    def test_mixed_on_different_host_when_host_already_exists(self):
        """ Register Host A with W1 and W2 when the following hosts are already registered:
        A with ports Q1 and Q2 and host B with with W1 and Q3 """
        self.register_host(self.hostname, [self.other_port1, self.other_port2])
        self.register_host("foo", [self.our_port1, self.other_port3])
        self.run_register()
        self.assert_host_exists("foo", [self.other_port3])
        self.assert_host_exists(self.hostname, [self.other_port1, self.other_port2])
        self.assert_host_exists(self.hostname + "-1", self.our_ports)

    def test_mixed_on_different_host_when_host_has_one_port(self):
        """ Register Host A with W1 and W2 when the following hosts are already registered:
        A with ports W2 and host B with with W1 and Q1 """
        self.register_host(self.hostname, [self.our_port2])
        self.register_host("foo", [self.our_port1, self.other_port1])
        self.run_register()
        self.assert_host_exists("foo", [self.other_port1])
        self.assert_host_exists(self.hostname, self.our_ports)

    def test_registered_as_different_host_when_host_already_exists(self):
        """ Register Host A with W1 and W2 when the following hosts are already registered:
        A with ports Q1 and Q2 and host B with with W1 and W2 """
        self.register_host(self.hostname, [self.other_port1, self.other_port2])
        self.register_host("foo", [self.our_port1, self.our_port2])
        self.run_register()
        self.assert_host_deleted("foo")
        self.assert_host_exists(self.hostname, [self.other_port1, self.other_port2])
        self.assert_host_exists(self.hostname + "-1", self.our_ports)

    def test_mixed_on_different_host_with_volumes(self):
        """ Register Host A with W1 and W2 when Host B with ports W1 and Q1 is already registered
        and has mapped volumes """
        foo_host = self.register_host("foo", [self.our_port1, self.other_port1])
        volume = self.create_volume()
        foo_host.map_volume(volume)
        self.run_register()
        self.assert_host_exists("foo", [self.other_port1])
        self.assert_host_exists(self.hostname, self.our_ports)
        self.assertIn(volume, lun_dict(foo_host.get_luns()).values())
        matching_hosts = [host for host in self.system.hosts if host.get_name() == self.hostname]
        self.assertEqual(len(matching_hosts), 1)
        self.assertIn(volume, lun_dict(matching_hosts[0].get_luns()).values())

    def test_one_port_registered_on_different_host_with_volumes(self):
        """ Register Host A with W1 and W2 when Host B with ports W1 is already registered
        and has mapped volumes """
        # expected: Host B will be renamed and contain W1 and W2
        foo_host = self.register_host("foo", [self.our_port1])
        volume = self.create_volume()
        foo_host.map_volume(volume)
        self.run_register()
        self.assert_host_deleted("foo")
        self.assert_host_exists(self.hostname, self.our_ports)
        matching_hosts = [host for host in self.system.hosts if host.get_name() == self.hostname]
        self.assertEqual(len(matching_hosts), 1)
        self.assertIn(volume, lun_dict(matching_hosts[0].get_luns()).values())

    def test_volume_is_mapped_to_lun0(self):
        foo_host = self.register_host("foo", self.our_ports)
        volume = self.create_volume()
        foo_host.map_volume(volume, lun=0)
        self.run_register()
        self.assertNotEquals(foo_host.get_name(from_cache=False), 'foo')

try:
    import resource
    resource.setrlimit(resource.RLIMIT_NOFILE, (4096, 4096))
except (ValueError, ImportError):
    # not allowed to raise maximum limit
    raise SkipTest("not allowed to raise maximum limit")
