from unittest import SkipTest
from infi import unittest
from mock import patch
from ...cli.engine import run_system_register, run_system_list
from infi.run_as_root import RootPermissions
from infi.vendata.powertools.tests.mocks import single_system__registered
from infi.vendata.powertools.tests.mocks import single_system__unregistered
from infi.vendata.powertools.tests.mocks import three_system__unregistered
from infi.vendata.powertools.tests.mocks import four_systems__two_registered

# pylint: disable-all


class RunSystemRegisterTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(RunSystemRegisterTestCase, cls).setUpClass()
        if cls.should_skip():
            raise SkipTest("Base Class")

    def setUp(self):
        self._is_root_patch = getattr(patch, "object")(RootPermissions, "is_root")
        self._is_root = self._is_root_patch.start()
        super(RunSystemRegisterTestCase, self).setUp()

    def tearDown(self):
        self._is_root_patch.stop()
        super(RunSystemRegisterTestCase, self).tearDown()

    @classmethod
    def should_skip(cls):
        return True

    def _run_system_register(self, system="autodetect", hostname="autodetect", auto_fix=False, fix=False):
        return run_system_register({"--system": system, "--hostname": hostname, "--auto-fix": auto_fix, "--fix": fix})

    def test_register__one_system_attached__not_registered(self):
        self._run_system_register()

    def test_show__one_system_attached__not_registered(self):
        run_system_list({})

    def test_show__mix_of_systems(self):
        run_system_list({})

    def test_register__three_systems_attached__not_registered(self, system_serial=None):
        self._run_system_register(system=str(system_serial))


class RunSystemRegisterTestCaseWithSmock(RunSystemRegisterTestCase):
    @classmethod
    def should_skip(cls):
        return False

    def test_register__connectivity_objects_return_local_connectivity(self):
        from infi.storagemodel.connectivity import ConnectivityFactory
        mock = single_system__unregistered.setUp()
        with mock, patch.object(ConnectivityFactory, "get_fc_hctl_mappings", return_value={}):
            RunSystemRegisterTestCase.test_register__one_system_attached__not_registered(self)

    def test_register__one_system_attached__not_registered(self):
        mock = single_system__unregistered.setUp()
        with mock:
            RunSystemRegisterTestCase.test_register__one_system_attached__not_registered(self)

    def test_show__one_system_attached__not_registered(self):
        mock = single_system__unregistered.setUp()
        with mock:
            RunSystemRegisterTestCase.test_show__one_system_attached__not_registered(self)

    def test_show__mix_of_systems(self):
        mock = four_systems__two_registered.setUp()
        with mock:
            RunSystemRegisterTestCase.test_show__mix_of_systems(self)

    def test_register__three_systems_attached__not_registered(self, system_serial=None):
        mock = three_system__unregistered.setUp()
        system = mock.get_inventory().get_infiniboxes()[-1]
        system_serial = system.get_serial()
        with mock:
            host_count = mock.get_inventory().get_infiniboxes()[-1].hosts.count()
            RunSystemRegisterTestCase.test_register__three_systems_attached__not_registered(self, system_serial=system_serial)
            self.assertEqual(system.hosts.count(), host_count + 1)

    def test_register__one_system_attached_and_registered_rename(self):
        mock = single_system__registered.setUp()
        with mock:
            self._run_system_register(hostname="foo", auto_fix=True)

    def test_register__one_system_not_registered_metadata(self):
        from infi.vendata.powertools.__version__ import __version__ as powertools_version
        mock = single_system__unregistered.setUp()
        with mock:
            self._run_system_register()
            system = mock.get_inventory().get_infiniboxes()[-1]
            self.assertEquals(system.hosts.count(), 1)
            host = list(system.hosts)[0]
            metadata = host.get_all_metadata()
            self.assertGreater(len(metadata), 1)
            self.assertEquals(metadata['powertools_version'], powertools_version)

    def test_register_with_read_only_permissions(self):
        def _create_read_only_user(mock):
            box = mock.get_inventory().get_infiniboxes()[0]
            box.users.create(name='user', password='pass', role='ReadOnly', email='alias@domain.com')
            return "user", "pass"

        def _set_powertools_credentials(username, password):
            from infi.vendata.powertools.utils import auth
            from infi.credentials_store import Credentials
            auth.CREDENTIALS_STORE.set_default_credentials(Credentials(username, password))

        mock = single_system__unregistered.setUp()
        username, password = _create_read_only_user(mock)
        _set_powertools_credentials(username, password)
        self.addCleanup(_set_powertools_credentials, 'admin', '123456')

        with mock:
            self.assertEquals(self._run_system_register(), 1)


class RunSystemRegisterForRealTestCase(RunSystemRegisterTestCase):
    @classmethod
    def should_skip(cls):
        # TODO
        return True
        try:
            from infi.storagemodel import get_storage_model
            get_storage_model()
            return False
        except:
            pass
        return True


# os.pipe raises OSError("Too many file descriptors"
# workaround taken from http://bugs.python.org/issue11284
try:
    import resource
    resource.setrlimit(resource.RLIMIT_NOFILE, (4096, 4096))
except (ValueError, ImportError):
    # not allowed to raise maximum limit
    raise SkipTest("not allowed to raise maximum limit")
