# flake8: noqa
# This file is here to allow its reuse by vmware powertools.
# If we put it under /tests, it doesn't get included in the powertools egg.

from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By
from selenium.webdriver.support.select import Select
from infi.pyutils.lazy import cached_method
from time import sleep

from logging import getLogger
logger = getLogger('selenium_tests')


class DomObject(object):

    def __init__(self, elem):
        self.elem = elem
        self._verify_elem()
        self._load_children()

    def _verify_elem(self):
        """ A hook for making sure we've got the right DOM element """
        pass

    def _load_children(self):
        """ A hook for instantiating any child objects contained within this one """
        pass

    def select(self, css_selector):
        """ Find a list of child elements that match the CSS selector """
        ret = self.elem.find_elements_by_css_selector(css_selector)
        logger.info('select "%s": %d matches', css_selector, len(ret))
        return ret

    def select_one(self, css_selector):
        """ Find the first child element that matches the CSS selector, or None """
        ret = self.elem.find_element_by_css_selector(css_selector)
        logger.info('select_one "%s": %s', css_selector, 'found' if ret else 'no match')
        return ret

    def wait_for_element(self, css_selector):
        """
        Wait until there's an element that matches the CSS selector.
        Considers all elements on the page, not just children.
        """
        logger.info('wait_for_element "%s"', css_selector)
        WebDriverWait(self.elem.parent, 10).until(EC.presence_of_element_located((By.CSS_SELECTOR, css_selector)))

    def wait_for_visible_element(self, css_selector):
        """
        Wait until there's an element that matches the CSS selector, and it is visible to the user.
        Considers all elements on the page, not just children.
        """
        logger.info('wait_for_visible_element "%s"', css_selector)
        WebDriverWait(self.elem.parent, 10).until(EC.visibility_of_element_located((By.CSS_SELECTOR, css_selector)))

    def wait_until_gone(self, css_selector):
        """
        Wait until an element is no longer present, or invisible.
        Considers all elements on the page, not just children.
        """
        logger.info('wait_until_gone "%s"', css_selector)
        WebDriverWait(self.elem.parent, 10).until(EC.invisibility_of_element_located((By.CSS_SELECTOR, css_selector)))

    def set_checkbox_value(self, css_selector, checked):
        """
        Checks or unchecks a child checkbox that matches the given selector.
        """
        logger.info('set_checkbox_value "%s" to %s', css_selector, checked)
        elem = self.select_one(css_selector)
        if elem.is_selected() != checked:
            elem.click()

    def set_field_value(self, css_selector, value):
        """
        Sets the text of a child input/select that matches the given selector.
        """
        logger.info('set_field_value "%s" to "%s"', css_selector, value)
        elem = self.select_one(css_selector)
        if elem.tag_name == 'select':
            Select(elem).select_by_visible_text(value)
        else:
            elem.clear()
            elem.send_keys(value)


class DomPage(DomObject):
    def _get_dialogs(self, dialog_classes):
        for elem in self.select('.ui-dialog'):
            if elem.value_of_css_property('display') == 'block':
                title = DomObject(elem).select_one('.ui-dialog-title').text
                if title not in dialog_classes:
                    raise Exception('Unrecognized dialog: %s' % title)
                yield dialog_classes[title](elem)

    def wait_for_dialog(self):
        """ Wait for the next dialog to pop up, and return its matching DomObject """
        self.wait_for_visible_element('.ui-widget-overlay')
        dialogs = self.get_dialogs()
        if not dialogs:
            raise Exception('No visible dialog found')
        return dialogs[0]

    def get_dialogs(self):
        return list(self._get_dialogs(self.DIALOG_CLASSES))

    def close_all_dialogs(self):
        dialogs = self.get_dialogs()
        for dialog in reversed(dialogs):
            dialog.close()


class MainPage(DomPage):

    """ Represents the main PowerTools page """

    ACTIONS_BUTTON = '.header .actions-button'
    CHECK_READINESS = '.header .actions-menu > ul > li:first-child > a'

    def _verify_elem(self):
        assert self.elem.tag_name == 'body'

    def _load_children(self):
        self.infiniboxes = []
        for elem in self.select('.infinibox.root'):
            self.infiniboxes.append(InfiniBox(elem, self))

    def is_host_ready(self):
        span = self.select_one('span.unready')
        return span.value_of_css_property('display') == 'none'

    def check_host_readiness(self):
        self.select_one(MainPage.ACTIONS_BUTTON).click()
        self.select_one(MainPage.CHECK_READINESS).click()
        actions_dialog = self.wait_for_dialog()
        actions_dialog.run_and_dismiss()


class InfiniBox(DomObject):

    """ Represents one InfiniBox panel on the page """

    USERNAME         = '.infinibox-username'
    LOGIN_LINK       = '.set-credentials'
    REGISTER_BUTTON  = '.register-button'
    PROVISION_BUTTON = '.create_new_popup'

    def __init__(self, elem, main_page):
        super(InfiniBox, self).__init__(elem)
        self._main_page = main_page

    def _verify_elem(self):
        assert self.elem.get_attribute('class') == 'infinibox root'

    def _load_children(self):
        self.volumes = []
        if self.select_one('.inventory tbody td').get_attribute('class') != 'dataTables_empty':
            for elem in self.select('.inventory tbody tr'):
                self.volumes.append(Volume(elem, self))
        for v in self.volumes:
            print v

    def refresh_volumes(self):
        self._load_children()

    def is_logged_in(self):
        return self.select_one(InfiniBox.USERNAME).text != 'No credentials set'

    def login(self, username='admin', password='123456', remember=False, default=False):
        self.select_one(InfiniBox.LOGIN_LINK).click()
        login_dialog = self._main_page.wait_for_dialog()
        login_dialog.login(username, password, remember, default)
        assert self.is_logged_in()

    def auto_login(self):
        if not self.is_logged_in():
            self.login()

    def is_registered(self):
        return 'unregistered' not in self.select_one('div.information').get_attribute('class')

    def register(self):
        self.auto_login()
        button = self.select_one(InfiniBox.REGISTER_BUTTON)
        button.click()
        self.wait_for_visible_element(InfiniBox.PROVISION_BUTTON)

    def provision(self, name=None, size=10, unit='GB', thin=False, pool=None, lun=None, filesystem=None, path=None):
        """
        Create a new volume. Returns the Volume instance, or None in case the volume cannot be created.
        """
        if not self._main_page.is_host_ready():
            self._main_page.check_host_readiness()
        self.auto_login()
        button = self.select_one(InfiniBox.PROVISION_BUTTON)
        button.click()
        prov_dialog = self._main_page.wait_for_dialog()
        name = prov_dialog.provision(name, size, unit, thin, pool, lun, filesystem, path)
        next_step = self._main_page.wait_for_dialog()
        if isinstance(next_step, ErrorDialog):
            self._main_page.close_all_dialogs()
            return None
        next_step.run_and_dismiss()
        self.refresh_volumes()
        for volume in self.volumes:
            if volume.get_column_texts()[0] == name:
                return volume
        raise Exception('Cannot find the new volume "%s" in the list' % name)


class Volume(DomObject):

    ACTIONS_BUTTON = '.actions-button'

    def __init__(self, elem, box):
        super(Volume, self).__init__(elem)
        self._box = box

    def _verify_elem(self):
        assert self.elem.tag_name == 'tr'
        assert self.select_one('.actions') is not None

    @cached_method
    def get_column_texts(self):
        return [elem.text for elem in self.select('td')[:-1]]

    @cached_method
    def get_actions(self):
        return [elem.get_attribute('class').split()[0] for elem in self.select('.action-button')]

    def __str__(self):
        return 'Volume' + str(self.get_column_texts())

    def click_action(self, name):
        self.select_one(Volume.ACTIONS_BUTTON).click()
        self.select_one('a.' + name).click()

    def unmap(self):
        self.click_action('unmap')
        confirm_dialog = self._box._main_page.wait_for_dialog()
        confirm_dialog.confirm()
        self._box.refresh_volumes()

    def unmount(self):
        self.click_action('unmount')
        confirm_dialog = self._box._main_page.wait_for_dialog()
        confirm_dialog.confirm()
        self._box.refresh_volumes()

    def rename(self, new_name):
        self.click_action('rename')
        rename_dialog = self._box._main_page.wait_for_dialog()
        rename_dialog.rename(new_name)
        success = True
        dialogs = self._box._main_page.get_dialogs()
        if dialogs:
            self._box._main_page.close_all_dialogs()
            success = False
        self._box.refresh_volumes()
        return success


class DomDialog(DomObject):
    def close(self):
        """ cancel / dismiss dialog """
        self.select('.ui-dialog-buttonset .ui-button')[-1].click()


class LoginDialog(DomDialog):

    """ Represents a login dialog """

    CONTAINER = '#auth-modal'
    USERNAME  = '#auth-modal-username'
    PASSWORD  = '#auth-modal-password'
    REMEMBER  = '#auth-modal-remember'
    DEFAULT   = '#auth-modal-default'

    def _verify_elem(self):
        assert 'ui-dialog' in self.elem.get_attribute('class')
        assert self.select_one(LoginDialog.CONTAINER) is not None

    def login(self, username='admin', password='123456', remember=False, default=False):
        self.select_one(LoginDialog.USERNAME).send_keys(username)
        self.select_one(LoginDialog.PASSWORD).send_keys(password)
        if remember:
            self.select_one(LoginDialog.REMEMBER).click()
        if default:
            self.select_one(LoginDialog.DEFAULT).click()
        self.select_one('.ui-dialog-buttonset .ui-button').click()
        self.wait_until_gone(LoginDialog.CONTAINER)


class ProvisionDialog(DomDialog):

    """ Represents a volume provisioning dialog """

    CONTAINER   = '#create-vol-modal'
    NAME        = '#vol-modal-vol-name'
    SIZE        = '#vol-modal-size-text'
    UNIT        = '#vol-modal-size-unit'
    THIN        = '#vol-modal-thickthin'
    POOL        = '#vol-modal-select-pool'
    LUN         = '#vol-modal-lun'
    FS_CREATE   = '#vol-modal-create-fs'
    FS_TYPE     = '#vol-modal-select-fs'
    FS_MOUNT    = '#vol-modal-mount-fs'
    PATH        = '#vol-modal-mnt-text'

    def _verify_elem(self):
        assert 'ui-dialog' in self.elem.get_attribute('class')
        assert self.select_one(ProvisionDialog.CONTAINER) is not None

    def provision(self, name=None, size=10, unit='GB', thin=False, pool=None, lun=None, filesystem=None, path=None):
        """
        Fills the dialog and submits it (this does not create the volume yet). Returns the new volume's name.
        Pass filesystem=False to skip filesystem creation.
        Pass path=False to skip mounting.
        """
        # Name
        if name:
            self.set_field_value(ProvisionDialog.NAME, name)
        else:
            name = self.select_one(ProvisionDialog.NAME).get_attribute('value')
        # Size and units
        self.set_field_value(ProvisionDialog.SIZE, str(size))
        self.set_field_value(ProvisionDialog.UNIT, unit)
        # Thin provisioning
        self.set_checkbox_value(ProvisionDialog.THIN, thin)
        # Pool
        if pool:
            self.set_field_value(ProvisionDialog.POOL, pool)
        # Lun
        if lun:
            self.set_field_value(ProvisionDialog.LUN, str(lun))
        # Filesystem
        if filesystem is False:
            self.set_checkbox_value(ProvisionDialog.FS_CREATE, False)
        else:
            self.set_checkbox_value(ProvisionDialog.FS_CREATE, True)
            if filesystem:
                self.set_field_value(ProvisionDialog.FS_TYPE, filesystem)
        # Mount path
        if path is False:
            self.set_checkbox_value(ProvisionDialog.FS_MOUNT, False)
        else:
            self.set_checkbox_value(ProvisionDialog.FS_MOUNT, True)
            if path:
                self.set_field_value(ProvisionDialog.PATH, path)
        # Submit the form
        sleep(1)
        self.select_one('.ui-dialog-buttonset .ui-button').click()
        self.wait_until_gone(ProvisionDialog.CONTAINER)
        return name


class ActionsDialog(DomDialog):

    CONTAINER = '#actions-modal'

    def _verify_elem(self):
        assert 'ui-dialog' in self.elem.get_attribute('class')
        assert self.select_one(ActionsDialog.CONTAINER) is not None

    def run_and_dismiss(self):
        self.select('.ui-dialog-buttonset .ui-button')[-2].click()
        self.wait_until_gone('.spinner-ajax-loader')
        self.select('#actions-modal-done')[-1].click()
        self.wait_until_gone('#actions-modal')
        self.wait_until_gone('.blockUI')


class ConfirmDialog(DomDialog):

    CONTAINER = '#confirm-modal'

    def _verify_elem(self):
        assert 'ui-dialog' in self.elem.get_attribute('class')
        assert self.select_one(ConfirmDialog.CONTAINER) is not None

    def confirm(self):
        self.select('.ui-dialog-buttonset .ui-button')[-2].click()
        self.wait_until_gone('.blockUI')

    def cancel(self):
        self.select('.ui-dialog-buttonset .ui-button')[-1].click()


class ErrorDialog(DomDialog):

    CONTAINER = '#error-modal'

    def _verify_elem(self):
        assert 'ui-dialog' in self.elem.get_attribute('class')
        assert self.select_one(ErrorDialog.CONTAINER) is not None

    def dismiss(self):
        self.select('.ui-dialog-buttonset .ui-button')[-1].click()


class RenameDialog(DomDialog):

    CONTAINER = '#rename-modal'
    NAME = '#rename-modal-name'

    def _verify_elem(self):
        assert 'ui-dialog' in self.elem.get_attribute('class')
        assert self.select_one(RenameDialog.CONTAINER) is not None

    def rename(self, new_name):
        self.set_field_value(RenameDialog.NAME, new_name)
        self.select('.ui-dialog-buttonset .ui-button')[-2].click()
        self.wait_until_gone('.blockUI')

    def cancel(self):
        self.select('.ui-dialog-buttonset .ui-button')[-1].click()


# A mapping from dialog title to the appropriate dialog class
DomPage.DIALOG_CLASSES = {
    'SET CREDENTIALS':          LoginDialog,
    'AUTHENTICATION REQUIRED':  LoginDialog,
    'CREATE VOLUME':            ProvisionDialog,
    'CREATE VOLUME STEPS':      ActionsDialog,
    'HOST READINESS':           ActionsDialog,
    'CONFIRM':                  ConfirmDialog,
    'ERROR':                    ErrorDialog,
    'RENAME VOLUME':            RenameDialog
}
