from inspect import isclass
from infi.pyutils.contexts import contextmanager
from infi.gevent_utils.silent_greenlets import spawn, join, joinall
from gevent import sleep, queue
from gevent.event import Event, AsyncResult
from json import dumps as encode
from json import loads as decode
from logging import getLogger
from infi.traceback import traceback_context
from infi.exceptools import exceptools_context
import jsonpickle
from infi.vendata.powertools.cli.engine import suppress_warnings

logger = getLogger(__name__)
SHUTDOWN_REQUEST = dict(object="shutdown_event", method="set", args=tuple(), kwargs=dict())
MAX_PIPE_CHUNK_SIZE = 100 * 1024


class BlockingRpcException(Exception):
    pass


class TimeoutExpired(BlockingRpcException):
    pass


class WorkerAlreadyStartedException(BlockingRpcException):
    pass


class WorkerNotRunningException(BlockingRpcException):
    pass


class WorkerHasShutDownException(BlockingRpcException):
    pass


class WorkerDiedUnexpectedlyException(BlockingRpcException):
    pass


class WorkerStuckException(BlockingRpcException):
    """raised when worker remains running after terminating it"""
    pass


def try_except_wrapper(func):
    from infi.pyutils.decorators import wraps
    @wraps(func)
    def decorated_function(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except KeyboardInterrupt:
            pass
    return decorated_function


class CooperativeProxy(object):
    """Proxy object for handling non-cooperative objects"""
    def __init__(self, non_cooperative_class, init_args=None, init_kwargs=None, name=None):
        super(CooperativeProxy, self).__init__()
        assert isclass(non_cooperative_class)
        self._worker_class = non_cooperative_class
        self._worker_init_args = init_args or tuple()
        self._worker_init_kwargs = init_kwargs or dict()
        self._worker = None
        self._name = name or str(non_cooperative_class.__name__)

    def start_worker(self):
        from infi.vendata.powertools.utils import auth
        from gipc import pipe, start_process
        if self._worker is not None:
            raise WorkerAlreadyStartedException()
        pipe_context = pipe(duplex=True)
        pipes = pipe_context.__enter__()  # child, parent
        self._pipes = pipes  # we keep reference to the pipe so they won't get closed before the cooperative proxy
        self._pipe_context = pipe_context
        self._queue = queue.Queue(maxsize=1)  # this will cause queue.put() to block until there is room in the queue
        self._shutdown_event = Event()
        args = (self._worker_class, self._worker_init_args, self._worker_init_kwargs, pipes[0], self._name,
                auth.CREDENTIALS_STORE.get_temp_credentials_dict())
        self._worker = start_process(spawn_worker, args)
        self._worker.join = try_except_wrapper(self._worker.join)
        self._handler = spawn(self._worker_handler_wrapper(), self._shutdown_event, pipes[1])
        sleep(0)

    def _worker_handler_wrapper(self):
        def _func(shutdown_event, pipe):
            return self._worker_handler(shutdown_event, pipe)
        _func.__name__ = ".".join([self.__class__.__name__, "_worker_handler", self._worker_class.__name__])
        return _func

    def _get_chunked(self, pipe):
        result = ""
        while True:
            next_chunk = pipe.get()
            if next_chunk is StopIteration:
                return result
            result += next_chunk

    def _worker_handler(self, shutdown_event, pipe):
        while not shutdown_event.is_set():
            async_result, request = self._queue.get()
            try:
                logger.debug("{} sending rpc blocking request: {}".format(self, request))
                pipe.put(encode(request))
                result = decode(self._get_chunked(pipe))
                logger.debug("{} got response for {}".format(self, request))
                if result["code"] == "error":
                    try:
                        async_result.set_exception(jsonpickle.decode(result["result"]))
                    except:
                        async_result.set_exception(RuntimeError(result))
                        logger.debug("failed to decode json resonse", exc_info=True)

                else:
                    async_result.set(result["result"])
            except (OSError, EOFError):
                logger.error("worker has died unexpectedly")
                async_result.set_exception(WorkerDiedUnexpectedlyException())
                shutdown_event.set()

    def _cancel_pending_requests(self):
        while True:  # raise WrokerHasShutDown for all pending operations
            try:
                async_result, request = self._queue.get_nowait()
                async_result.set_exception(WorkerHasShutDownException())
            except queue.Empty:
                break

    def stop_worker(self, timeout=None, block=True):
        """attempts to close the worker nicely
        :param timeout: if timeout is not None, will attempt to kill/terminate the worker after timeout
        :raises `WorkerStuckException`: if worker is still running after we terminated it
        """
        if self._worker is None:
            return
        worker, handler = self._worker, self._handler  # stop accepting new requests
        self._worker, self._handler = None, None
        self._cancel_pending_requests()

        def _stop_worker():
            self._shutdown_event.set()  # shuts down the worker handler
            self._queue.put((AsyncResult(), SHUTDOWN_REQUEST))  # shuts down the worker process
            wait_for_worker = lambda: worker.join(timeout)
            joinall([spawn(wait_for_worker), handler], timeout=timeout)
            if not handler.ready() or worker.is_alive():
                handler.kill()
                worker.terminate()
                joinall([spawn(wait_for_worker), handler], timeout=1)  # is one second enough?
            self._pipe_context.__exit__(None, None, None)

        greenlet = spawn(_stop_worker)
        if block:
            join(greenlet)
            if not handler.ready() or worker.is_alive():
                raise WorkerStuckException()
        else:
            return greenlet

    def __call__(self, method_name, *args, **kwargs):
        from gevent import Timeout
        from uuid import uuid4
        if self._worker is None:
            raise WorkerNotRunningException()
        kwargs['__rpc_uuid'] = str(uuid4())
        request = dict(object="worker", method=method_name, args=args, kwargs=kwargs)
        async_result = AsyncResult()
        logger.debug("sending blocking rpc request {!r}".format(request))
        self._queue.put((async_result, request))
        try:
            result = async_result.get()
            logger.debug("got blocking rpc result for request {!r}".format(request))
            return result
        except Timeout:
            raise TimeoutExpired()

    def __getattr__(self, method_name):
        attr = lambda *args, **kargs: self(method_name, *args, **kargs)
        attr.__name__ = method_name
        attr.__doc__ = self._worker_class.__doc__
        return attr

    def __dir__(self):
        return dir(super(CooperativeProxy, self)) + dir(self._worker_class)

    def __enter__(self):
        self.start_worker()

    def __exit__(self, type, value, tb):
        self.stop_worker()


@contextmanager
def setup_teardown_context(worker):
    nothing = lambda: None
    getattr(worker, "setup", nothing)()
    try:
        yield
    finally:
        getattr(worker, "teardown", nothing)()


def _setup_signal_null_handlers():
    from signal import SIGINT
    from gevent import signal
    signal(SIGINT, lambda: None)


def send_chunked(pipe, obj):
    # gevent doesn't handle large chunks well, see IZBOX-1605
    data = encode(obj)
    while len(data) > 0:
        chunk_size = min(len(data), MAX_PIPE_CHUNK_SIZE)
        pipe.put(data[:chunk_size])
        data = data[chunk_size:]
    pipe.put(StopIteration)


def _worker(cls, init_args, init_kwargs, pipe, logger):
    shutdown_event = Event()
    worker = cls(*init_args, **init_kwargs)
    with pipe, setup_teardown_context(worker):
        while not shutdown_event.is_set():
            request, response = None, None  # clean-up the response from previous iteration from the current frame
            try:
                request = decode(pipe.get())
            except EOFError:
                shutdown_event.set()
                continue
            try:
                _obj = locals().get(request['object'])
                method = getattr(_obj, request['method'])
                logger.debug("{} received blocking rpc request: {}".format(worker, request))
                args, kwargs = request['args'], request['kwargs']
                kwargs.pop('__rpc_uuid', None)
                response = dict(code="success", result=method(*args, **kwargs))
                logger.debug("{} sending blocking rpc response to {}".format(worker, request))
            except Exception as error:
                msg = "an unhandled exception ocurred during non-cooperative call to {!r} with {!r} {!r}"
                logger.exception(msg.format(request['method'], request['args'], request['kwargs']))
                response = dict(code="error", result=jsonpickle.encode(error))
            send_chunked(pipe, response)


def silence_requests_warning():
    import warnings
    with warnings.catch_warnings():
        import infinisdk  # noqa


@suppress_warnings
def spawn_worker(cls, init_args, init_kwargs, pipe, name, temp_credentials_dict):
    # It's very important that the entire function is wrapped with try/except, because if a gipc process
    # has an unhandled exception it will print to stderr and not log the exception - very bad
    try:
        from infi.vendata.powertools.utils import auth
        from infi.vendata.powertools.utils.logging import redirect_logbook
        from .logging import root, DebugLogging, DEBUG
        root.setLevel(DEBUG)
        silence_requests_warning()
        redirect_logbook()
        auth.CREDENTIALS_STORE.replace_temp_credentials_dict(temp_credentials_dict)
        debug = DebugLogging("infinihost_" + name)
        debug.reset()
        debug.set()
        logger = getLogger("infinihost_" + name)
        try:
            _setup_signal_null_handlers()
            with exceptools_context(), traceback_context():
                _worker(cls, init_args, init_kwargs, pipe, logger)
        except:
            logger.exception("Caught unhandled exception in worker process!")
    except:
        # should only happen if logging setup failed
        pass
