from infi.pyutils.decorators import wraps
from infi.pyutils.contexts import contextmanager
from os import environ, path

SMOCK_KEY = "SMOCK_ENVIRONMENT"

@contextmanager
def mock_root_permissions():  # pragma: no cover
    from mock import patch
    from infi.run_as_root import RootPermissions
    with patch.object(RootPermissions, "is_root") as is_root, patch('infi.run_as_root.ensure_root'):
        is_root.return_value = True
        yield

@contextmanager
def mock_speedtest():
    from mock import patch
    from infi.vendata.powertools.volume_speedtest.engine import PerformanceEngine
    def create_io_processes_mock_thread(device_path, process_count, block_size, ipc):
        for i in range(process_count):
            ipc.thread_ready_queue.put("Thread ready")
        ipc.start_event.wait()
        while not ipc.close_event.is_set():
            ipc.io_bytes_queue.put(block_size)
    def create_io_processes_mock(inst):
        import multiprocessing
        proc = multiprocessing.Process(target=create_io_processes_mock_thread,
                                args=(inst.device_path,
                                      inst.process_count,
                                      inst.block_size,
                                      inst.ipc))
        proc.start()
        inst.processes = [proc]
    with patch.object(PerformanceEngine, "_create_io_processes", create_io_processes_mock):
        yield

@contextmanager
def mock_settings_check():
    from infi.vendata.powertools.settings.runner import DemoRunner, PassRunner
    from infi.vendata.powertools.cli import engine
    from infi.vendata.powertools.gui.webapp.plugin.views.host import readiness
    from mock import patch
    import sys
    if hasattr(smock_in_use, 'runner_class'):
        runner_class = getattr(smock_in_use, 'runner_class')
    elif "gui" in sys.argv[0]:
        runner_class = PassRunner
    else:
        runner_class = DemoRunner
    with patch.object(engine, "ReadinessRunner", new=runner_class):
        with patch.object(readiness, "ReadinessRunner", new=runner_class):
            yield


@contextmanager
def mock_provisioning_check():
    from infi.vendata.powertools.cli import engine
    from mock import patch
    with patch.object(engine, "is_filesystem_support_disabled") as is_filesystem_support_disabled:
        is_filesystem_support_disabled.return_value = False
        yield


@contextmanager
def mock_mounts():
    from mock import patch
    from infi.vendata.powertools.volume_provision import engine as volume_provision
    from infi.vendata.powertools.volume_unmap import engine as volume_unmap
    from infi.vendata.powertools.cli import validate
    with patch.object(volume_provision, "makedirs"), \
         patch.object(volume_unmap, "rmtree"), \
         patch.object(validate, "rmdir"), \
         patch.object(validate, "mkdir"):
        yield

@contextmanager
def mock_rpc_blocking():
    from mock import patch
    from infi.vendata.powertools.volume_provision import engine as volume_provision
    with patch.object(volume_provision, "make_blocking", side_effect=lambda task: task):
        yield


@contextmanager
def mock_platform_string():
    from mock import patch
    from infi import os_info, execute
    with patch.object(os_info, 'get_platform_string', return_value='aix-7.1-powerpc'):
        with patch.object(execute, 'execute_assert_success'):
            yield


@contextmanager
def mock_get_reported_luns():
    from mock import patch
    from infi.vendata.powertools.settings.cases import devices

    def side_effect(device):
        from infinisdk import InfiniBox
        host_id = device.get_vendor().get_host_id()
        if host_id in (0, -1):
            return [0]
        system = InfiniBox(device.get_vendor().get_management_address(), auth=("admin", "123456"))
        luns = [int(lun) for lun in system.hosts.get(id=host_id).get_luns()]
        return list(set([0]).union(set(luns)))

    with patch.object(devices, "get_reported_luns", side_effect=side_effect):
        yield

@contextmanager
def mock_environment():  # pragma: no cover
    from colorama import Fore, init
    init()
    if environ.get(SMOCK_KEY) is None:
        yield
        return
    setup_filepath = path.abspath(environ.get(SMOCK_KEY))
    _globals, _locals = {}, {}
    execfile(setup_filepath, _globals, _locals)
    mock = _locals['setUp']()
    mock.start()
    print Fore.RED + "Mock started" + Fore.RESET
    try:
        yield mock
    finally:
        print Fore.RED + "Mock ended" + Fore.RESET
        mock.end()

global smock_in_use  # pylint: disable=W0604
smock_in_use = None

@contextmanager
def mock_all():
    global smock_in_use
    if smock_in_use:
        yield
        return
    with mock_environment() as mock, \
         mock_root_permissions(), \
         mock_speedtest(), \
         mock_mounts(), \
         mock_get_reported_luns(), \
         mock_platform_string(), \
         mock_rpc_blocking():
        smock_in_use = mock
        with mock_settings_check(), mock_provisioning_check():
            try:
                yield
            finally:
                smock_in_use = None

def consider_mock_environment(func):  # pragma: no cover
    @wraps(func)
    def _callable(*args, **kwargs):
        use_mock = bool(environ.get(SMOCK_KEY, False))
        if not use_mock:
            return func(*args, **kwargs)
        with mock_all():
            return func(*args, **kwargs)
    return _callable
