from os.path import exists, join, dirname

import logging
logger = logging.getLogger(__name__)

SIZE = "64x64"
API_SERVER = "http://chart.apis.google.com"
TEXT_TO_ENCODE = "{}"
API_URI = "chart?cht=qr&chs={}&chld=L|0&chl={}".format(SIZE, TEXT_TO_ENCODE)
GENERATE_URL = "{server}/{api_uri}".format(server=API_SERVER, api_uri=API_URI)

CACHE_FILE = join(dirname(__file__), "qrcode_dict.json")

class QRCodePool(object):
    def __init__(self, cache_filepath=CACHE_FILE):
        self._path = cache_filepath
        self._codes = {}
        self._load_from_disk()

    def _generate_png_file(self, key):
        import urllib
        import urllib2
        url = GENERATE_URL.format(urllib.quote(key))
        logger.info("Fetcing QRCode from: {}".format(url))
        result = urllib2.urlopen(url).read()
        return result

    def _strip_pixels_from_png(self, bytes):  # pylint: disable=0622
        from png import Reader
        obj = Reader(bytes=bytes)
        width, height, pixels, metadata = obj.read()
        logger.info("QRCode size is {}x{}".format(width, height))
        return [array for array in pixels]

    def _strip_usable_pixels(self, src):
        """The png module returns a list of byte arrays, each byte array represents a line.
        In each array, each three consecutive numbers (0,0,0 or 255,255,255) represent a pixel.
        Every 2x2 pixels represent the smallest square in png."""
        return [line.tolist()[::6] for line in src[::2]]

    def _stripped_pixels_to_colored_string(self, pixels):
        from colorama import Back
        res = '\n'.join(pixels)
        res = res.replace('1', Back.WHITE + " " + Back.RESET)
        res = res.replace('0', Back.BLACK + " " + Back.RESET)
        return res

    def _add_white_frame(self, pixels):
        horizon = [255, ] * (len(pixels[0]) + 2)
        vertical = [255, ]
        return [horizon, ] + [vertical + line + vertical for line in pixels] + [horizon, ]

    def _encode(self, code):
        """ Convert each line that contains pixels e.g. [255, 255, 0, 0...] to string e.g. '1100...' """
        return [''.join(['1' if pixel is 255 else '0' for pixel in line]) for line in code]

    def _generate(self, key):
        bytes = self._generate_png_file(key)  # pylint: disable=0622
        pixels = self._strip_pixels_from_png(bytes)
        stripped = self._strip_usable_pixels(pixels)
        code = self._add_white_frame(stripped)
        return [self._encode(code), bytes.encode("base64")]

    def add(self, key):
        self._codes[key] = self._generate(key)

    def _load_from_disk(self):
        import json
        if exists(self._path):
            with open(self._path, "r") as fd:
                try:
                    content = fd.read()
                    self._codes.update(json.loads(content))
                except ValueError:
                    logger.warning("{!r} is not a valid JSON encoding".format(content))

    def _dump_to_disk(self):
        import json
        with open(self._path, "w") as fd:
            json.dump(self._codes, fd, indent=4)

    def get(self, key):
        if key not in self._codes:
            raise KeyError()
        return self._stripped_pixels_to_colored_string(self._codes[key][0])

    def get_image_data(self, key):
        if key not in self._codes:
            raise KeyError()
        return self._codes[key][1].decode("base64")

    def save(self):
        self._dump_to_disk()

    def get_keys(self):
        return self._codes.keys()

    def update_all(self):
        for key in self._codes.keys():
            self.add(key)
