from progress_bar import progress_bar_context
from unittest.runner import _WritelnDecorator
from infi.pyutils.decorators import wraps
from ..utils import init_colors
from logging import getLogger
from os import environ
import colorama
import sys

logger = getLogger(__name__)

class Task(object):  # pragma: no cover
    """
    Base class for discrete tasks that can be run inside a TaskSequence.
    """
    def __init__(self, args):
        super(Task, self).__init__()
        self.args = args

    def get_description(self):
        raise NotImplementedError()

    def get_workout(self):
        return []

    def run(self):
        raise NotImplementedError()

def blocking_run(func):
    @wraps(func)
    def decorated(*args, **kwargs):
        from ..utils import blocking
        self = func.im_self
        proxy = blocking.CooperativeProxy(self.__class__, (self.args,))
        with proxy:
            result = proxy.run()
            if isinstance(result, dict):
                self.args.update(result)
            return result
    return decorated

def make_blocking(task):
    if not int(environ.get('HPT_CANCEL_MAKE_BLOCKING_TASKS', '0')):
        task.run = blocking_run(task.run)
    return task

class TaskSequence(object):
    """
    A utility for composing a complex operation using a sequence of discrete tasks.
    """
    def __init__(self, tasks):
        self.tasks = list(tasks)
        self.completed_tasks = []

    def get_description(self):
        """
        Get the cumulative description of all tasks in this sequence (list of strings).
        """
        return [task.get_description() for task in self.tasks]

    def get_workout(self):
        """
        Get the cumulative workout of all tasks in this sequence (list of strings).
        """
        workout = []
        for task in self.tasks:
            workout += task.get_workout()
        return workout

    def format_workout(self):
        """
        Get a string listing of the workouts of this task sequence.
        """
        msg = ""
        for task_msg in self.get_workout():
            msg += "* {}\n".format(task_msg)
        return msg

    def done(self):
        """
        Returns True if there are no more tasks to perform.
        """
        return len(self.tasks) == len(self.completed_tasks)

    def next(self):
        """
        Returns the next task to run, or None if all tasks were executed already.
        """
        return None if self.done() else self.tasks[len(self.completed_tasks)]

    def step(self):
        """
        Runs the next task. If the task raises an exception, it isn't caught here.
        """
        task = self.next()
        if task:
            task.run()
            self.completed_tasks.append(task)

    def run(self, stream=None):
        """
        Runs the sequence of tasks, printing the progress.
        """
        if stream is None:
            stream = _WritelnDecorator(sys.stdout)
        stream.flush()
        init_colors()
        while not self.done():
            task = self.next()
            stream.write("{} ... ".format(task.get_description()))
            stream.flush()
            try:
                with progress_bar_context():
                    self.step()
            except BaseException:
                print colorama.Fore.RED + "ERROR" + colorama.Fore.RESET
                raise
            print colorama.Fore.GREEN + "OK" + colorama.Fore.RESET
        stream.flush()
