import sys
from unittest.runner import _WritelnDecorator
from ..utils import round_to_block_size
from ..utils.tasks import Task, TaskSequence
from ..utils.queries import get_infinisdk_host_and_volume_by_device
from ..cli.validate import get_storagemodel_device, convert_size
from infi.storagemodel.base.partition import MBRPartitionTable
try:
    from infi.storagemodel.base.partition import GUIDPartitionTable
except:
    from infi.storagemodel.base.partition import GPTPartitionTable as GUIDPartitionTable
from capacity import byte, TiB
from munch import Munch
import platform
from logging import getLogger

logger = getLogger(__name__)


class ResizeError(Exception):
    pass


class ResizeTask(Task):

    def _verb(self):
        return 'Shrink' if self.args.newsize < self.args.oldsize else 'Expand'

    def _size(self):
        return str(self.args.newsize).replace('*', ' ')

    def _type(self):
        raise NotImplementedError()

    def get_workout(self):
        return ['{} the {} to {}'.format(self._verb(), self._type(), self._size())]

    def get_description(self):
        return '{}ing the {} to {}'.format(self._verb(), self._type(), self._size())


class ResizeFilesystem(ResizeTask):

    def _type(self):
        return 'filesystem'

    def run(self):
        # the partition size ia a bit less than the volume size, we need to take it into account
        diff = self.args.newsize - self.args.oldsize
        self._resize_filesystem(self.args.device, self._get_partition_size(self.args.device) + diff)

    def _get_partition_size(self, device):
        partition_table = device.get_disk_drive().get_partition_table()
        return partition_table.get_partitions()[-1].get_size_in_bytes() * byte

    def _resize_filesystem(self, device, size):
        partition_table = device.get_disk_drive().get_partition_table()
        partition = partition_table.get_partitions()[-1]
        partition.resize(size // byte)


class ResizeVolume(ResizeTask):

    def _type(self):
        return 'volume'

    def run(self):
        self._resize_volume(self.args.device, self.args.newsize)

    def _resize_volume(self, device, size):
        size = round_to_block_size(size // byte) * byte
        host, volume = get_infinisdk_host_and_volume_by_device(device)
        volume.update_size(size)


class Rescan(Task):

    def get_description(self):
        return 'Rescanning'

    def run(self):
        from ..rescan.engine import do_rescan
        do_rescan(disable_print=True)


def check_preconditions(args):
    from infi.vendata.powertools.cli.validate import validate_filesystem_size
    from infi.vendata.powertools.utils.queries import is_device_read_only
    if platform.system() != 'Windows':
        raise ResizeError('Resize is supported only on Microsoft Windows')
    if is_device_read_only(args.device):
        raise ResizeError('Device is read only')
    if args.newsize < args.oldsize:
        raise ResizeError('Volume size cannot be decreased, only increased')
    partition_table = args.device.get_disk_drive().get_partition_table()
    if not (isinstance(partition_table, MBRPartitionTable) or isinstance(partition_table, GUIDPartitionTable)):
        raise ResizeError('Resize is not supported for this type of partition table')
    if isinstance(partition_table, MBRPartitionTable) and args.newsize > 2*TiB:
        raise ResizeError('Resize to more than 2 TiB is not supported for an MBR partition table')
    partitions = partition_table.get_partitions()
    if len(partitions) != 1:
        raise ResizeError('Resize is not supported for drives with no partitions or more than one partition')
    validation_result = validate_filesystem_size("NTFS", args.newsize)
    if validation_result:
        raise ResizeError(validation_result)


def convert_arguments(arguments):
    device = get_storagemodel_device(arguments)
    oldsize = device.get_size_in_bytes() * byte
    newsize = convert_size(arguments['<size>'])
    if arguments.get('--expand-by'):
        newsize = oldsize + newsize
    return Munch(device=device, oldsize=oldsize, newsize=newsize)


def load_tasks(args):
    if args.newsize < args.oldsize:
        return TaskSequence([ResizeFilesystem(args),
                             ResizeVolume(args),
                             Rescan(args)])
    else:
        return TaskSequence([ResizeVolume(args),
                             Rescan(args),
                             ResizeFilesystem(args)])


def get_workout(arguments):
    args = convert_arguments(arguments)
    msg = "This tool is going to:\n" + load_tasks(args).format_workout()
    return msg


def resize(arguments, stream=_WritelnDecorator(sys.stdout)):
    args = convert_arguments(arguments)
    load_tasks(args).run(stream)
    stream.writeln("="*79)
    stream.flush()
    print "Volume and filesystem were resized."


def in_check_preconditions(arguments):
    try:
        args = convert_arguments(arguments)
        check_preconditions(args)
    except ResizeError as error:
        print "ERROR: {}".format(error.message)
        return 2
    return 0


def in_resize(arguments):
    from infinisdk.core.exceptions import APICommandFailed
    try:
        resize(arguments)
        return 0
    except APICommandFailed as exc:
        try:
            print "ERROR: {}".format(exc.message)
            return 3
        except:
            pass
        # couldn't parse the infinisdk error message to display somehting sensible to the user
        # reraise original infinisdk traceback and let debug_program catch this as a generic error
        raise
