import logging
import os
from infi.storagemodel.vendor.infinidat.predicates import InfinidatVolumeDoesNotExist
from shutil import rmtree

logger = logging.getLogger(__name__)

class UnmountException(Exception):
    pass


class UnmapVolume(object):
    def __init__(self, device):
        super(UnmapVolume, self).__init__()
        self._device = device

    def _umount_partition(self, access_path_to_mount_point, partition):
        import os
        from infi.storagemodel import get_storage_model
        access_path = partition.get_block_access_path()
        mount_point = access_path_to_mount_point.get(access_path, None)
        if mount_point is None:
            return
        def remove_mount():
            filesystem = partition.get_current_filesystem()
            filesystem.unmount(access_path, mount_point)
            if os.name != 'nt':
                rmtree(mount_point)
        def remove_persistent_mount():
            repository = get_storage_model().get_mount_repository()
            for persistent_mount in repository.get_all_persistent_mounts():
                if persistent_mount.get_mount_point().lower() == mount_point.lower():
                    repository.remove_persistent_mountpoint(persistent_mount)
                    break
        remove_mount()
        remove_persistent_mount()

    def unmount_device(self):
        from infi.storagemodel import get_storage_model
        from infi.storagemodel.base.disk import NoPartitionTable
        from infi.parted import InvalidPartitionTable
        from infi.storagemodel.base.multipath import MultipathBlockDevice
        if not isinstance(self._device, MultipathBlockDevice):
            return
        mounts = get_storage_model().get_mount_manager().get_mounts()
        access_path_to_mount_point = dict((mount.get_block_access_path(), mount.get_mount_point()) for mount in mounts)
        partitions = []
        try:
            partitions = list(self._device.get_disk_drive().get_partition_table().get_partitions())
        except (InvalidPartitionTable, NoPartitionTable) as err:
            logger.debug("No partitions found for device {}, reason: {}".format(self._device, err))
        for partition in partitions:
            self._umount_partition(access_path_to_mount_point, partition)

    def unmap_from_host(self):
        from ..utils.queries import get_infinisdk_host_and_volume_by_device
        from infi.vendata.powertools.cli.engine import is_filesystem_support_disabled
        from infi.vendata.powertools.utils.queries import is_device_read_only
        self.volume_id = self._device.get_vendor().get_volume_id()
        self.system_serial = self._device.get_vendor().get_system_serial()
        # Windows doesn't handle read only devices
        if not (is_filesystem_support_disabled() or os.name == 'nt' and is_device_read_only(self._device)):
            self.unmount_device()
        host, volume = get_infinisdk_host_and_volume_by_device(self._device)
        logger.info("Unmapping volume {}".format(volume.get_name()))
        for lun_mapping in host.get_luns():
            if lun_mapping.volume == volume:
                if lun_mapping.is_clustered():
                    raise UnmountException('The volume is cluster-mapped, it cannot be unmounted from this host')
                host.unmap_volume(volume=volume)

def get_workout(arguments):
    from infi.vendata.powertools.cli.engine import is_filesystem_support_disabled
    if is_filesystem_support_disabled():
        return "This tool will unmap the selected device from this host."
    return "This tool will unmount the selected device, and unmap it from this host."

def unmap(device):
    from ..utils import rescan_storagemodel
    operation = UnmapVolume(device)
    operation.unmap_from_host()
    rescan_storagemodel(InfinidatVolumeDoesNotExist(operation.system_serial, operation.volume_id))

def unmount(device):
    from infi.storagemodel import get_storage_model
    from infi.storagemodel.errors import UnmountFailedDeviceIsBusy
    operation = UnmapVolume(device)
    try:
        operation.unmount_device()
    except (UnmountException, UnmountFailedDeviceIsBusy) as exc:
        print "ERROR: {}".format(exc)
        return 1
    get_storage_model().refresh()
    return 0

def delete_device(device):
    from infinisdk.core.exceptions import APICommandFailed
    from infi.storagemodel.errors import UnmountFailedDeviceIsBusy
    from ..utils.queries import get_infinisdk_host_and_volume_by_device
    infinisdk_host, infinisdk_device = get_infinisdk_host_and_volume_by_device(device)
    try:
        unmap(device)
        infinisdk_device.delete()
    except APICommandFailed as exc:
        print "ERROR: {}".format(exc.message)
        return 1
    except (UnmountException, UnmountFailedDeviceIsBusy) as exc:
        print "ERROR: {}".format(exc)
        return 1
    return 0

def in_unmap(device):
    from infinisdk.core.exceptions import APICommandFailed
    from infi.storagemodel.errors import UnmountFailedDeviceIsBusy
    try:
        unmap(device)
        return 0
    except APICommandFailed as exc:
        print "ERROR: {}".format(exc.message)
        return 1
    except (UnmountException, UnmountFailedDeviceIsBusy) as exc:
        print "ERROR: {}".format(exc)
        return 1
